"""Abstract game rule."""

import abc


class AbstractRule(abc.ABC):
    """Abstract game rule class.
    
    All game rules must inherit from this class.
    """

    def reset(self, state, meta_state):
        """Reset rule at beginning of every trial.

        This method can be used to reset any attributes of the rule that serve
        as memory within trials.
        
        Args:
            state: OrderedDict of iterables of sprites. Environment state.
            meta_state: meta_state of environment.
        """
        pass
    
    @abc.abstractmethod
    def step(self, state, meta_state):
        """Apply rule to the environment state.

        This method can in-place modify the state however it likes.
        
        Args:
            state: OrderedDict of iterables of sprites. Environment state.
            meta_state: meta_state of environment.
        """
        pass



"""Rules that modify sprites."""

from . import abstract_rule
import itertools
import numpy as np


class ModifySprites(abstract_rule.AbstractRule):
    """Modify sprites in a layer or set of layers.

    A filter can be applied to modify only sprites within the layer of interest
    that satisfy some condition.
    """

    def __init__(self, layers, modifier, sample_one=False, filter_fn=None):
        """Constructor.

        Args:
            layers: String or iterable of strings. Must be a key (or keys) in
                the environment state. Layer(s) in which sprites are modified.
            modifier: Function taking in a sprite and modifying it in place.
            sample_one: Bool. Whether to sample one sprite to modify if multiple
                satisfy filter_fn at a given step.
            filter_fn: Optional filter function. If specified must take in a
                sprite and return a bool saying whether to consider modifying
                that sprite.
        """
        if isinstance(layers, str):
            layers = [layers]
        self._layers = layers
        self._modifier = modifier
        self._sample_one = sample_one
        self._filter_fn = filter_fn

    def step(self, state, meta_state):
        """Apply rule to state."""
        del meta_state
        
        sprites_to_modify = [s for k in self._layers for s in state[k]]

        if self._filter_fn:
            sprites_to_modify = list(
                filter(self._filter_fn, sprites_to_modify))

        if not sprites_to_modify:
            return

        if self._sample_one:
            sprites_to_modify = [np.random.choice(sprites_to_modify)]

        for sprite in sprites_to_modify:
            self._modifier(sprite)



"""Environment wrapper class for logging episodes.

This can be used to record data from a subject playing the task. See
../../moog_demos/restore_logged_data.py for an example of how to read log files.

Note: This logger records everything about the environment, which can be a lot
of data (depending on the task). If you plan to use this at scale for recording
subjects' or agents' behavior, we recommend forking this and modifying it to
only log the data that you need to do analyses for your specific task. For
example you may not want to log the positions/velocities of static sprites
(e.g. walls), or may not want to log all the attributes of sprites every
timestep (e.g. if you know that the colors of the sprites don't change in your
task).
"""

import copy
from datetime import datetime
import json
import logging
import numpy as np
import os
import time

from moog import env_wrappers
from moog import sprite

# This is the number of numerals in filenames. Since there is one file per
# episode, you should pick _FILENAME_ZFILL large enough that the number of
# episodes in your dataset is less than 10^_FILENAME_ZFILL.
_FILENAME_ZFILL = 5


class VertexLogging():
    NEVER = 'NEVER'
    ALWAYS = 'ALWAYS'
    WHEN_NECESSARY = 'WHEN_NECESSARY'


def _serialize(x):
    """Serialize a value x.

    This is used to serialize sprite attributes, actions, and meta_state so that
    they are json-writable.

    Specifically, numpy arrays are not JSON serializable, so we must convert
    numpy arrays to lists. This function is recursive to handle nestings inside
    of lists/tuples/dictionaries.

    Args:
        x: Value to serialize.

    Returns:
        Serialized value that can be JSON dumped.
    """
    if isinstance(x, np.ndarray):
        return x.tolist()
    elif isinstance(x, (np.float32, np.float64)):
        return float(x)
    elif isinstance(x, (np.int32, np.int64)):
        return int(x)
    elif isinstance(x, list):
        return [_serialize(a) for a in x]
    elif isinstance(x, tuple):
        return tuple([_serialize(a) for a in x])
    elif isinstance(x, dict):
        return {k: _serialize(v) for k, v in x.items()}
    else:
        return x


class LoggingEnvironment(env_wrappers.AbstractEnvironmentWrapper):
    """Environment class for logging timesteps.

    This logger produces a description of the log in 'description.txt' of
    log_dir, so please refer to that for a detailed account of the structure of
    the logs.
    """

    def __init__(self, environment, log_dir='logs',
                 log_vertices='WHEN_NECESSARY'):
        """Constructor.

        Args:
            environment: Instance of ../moog/environment.Environment.
            log_dir: String. Log directory relative to working directory.
            log_vertices: String. Of the following options:
                * 'NEVER'. In this case, never log sprite vertices.
                * 'WHEN_NECESSARY'. In this case, log sprite vertices when a
                    sprite has either just appeared or just changed shape. In
                    this way, the vertices of a sprite can always be inferred
                    from the current position/angle/aspect_ratio and the
                    vertices that were logged for that sprite (identifiable by
                    its id) the last time its vertices were logged.
                * 'ALWAYS'. Log vertices for all sprites every timestep.
        """
        super(LoggingEnvironment, self).__init__(environment)

        # Make sure log_vertices is a valid value
        if not hasattr(VertexLogging, log_vertices):
            raise ValueError('log_vertices is {} but must be in VertexLogging '
                             'values'.format(log_vertices))
        self._log_vertices = log_vertices

        # Set the logging directory
        now_str = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
        if log_dir[0] == '/':
            log_dir = os.path.join(log_dir, now_str)
        else:
            log_dir = os.path.join(os.getcwd(), log_dir, now_str)
        os.makedirs(log_dir)
        self._log_dir = log_dir

        # These are the attributes that we'll log
        self._attributes = list(sprite.Sprite.FACTOR_NAMES) + ['id']

        # Log attribute list
        attributes_filename = os.path.join(self._log_dir, 'attributes.txt')
        logging.info('Logging attribute list {} to {}.'.format(
            self._attributes, attributes_filename))
        with open(attributes_filename, 'w') as f:
            json.dump(self._attributes, f)

        # Log description
        self._log_description()

        # Initialize self._episode_log
        self._episode_count = 0
        self._episode_log = []

    def _log_description(self):
        """Log a description of the data to a description.txt file."""
        description_filename = os.path.join(self._log_dir, 'description.txt')
        logging.info('Logging description to {}.'.format(description_filename))
        description = (
            'Each numerical file in this directory is an episode of the task. '
            'Each such file contains a json-serialized list, each element of '
            'which represents an environment step in the episode. Each step is '
            'a list of four elements, [[`time`, time], [`reward`, reward], '
            '[`step_type`, step_type], [`action`, action], [`meta_state`, '
            'meta_state`], state].'
            '\n\n'
            '\n\n'
            'time is a timestamp of the timestep.'
            '\n\n'
            '\n\n'
            'reward contains the value of the reward at that step.'
            '\n\n'
            '\n\n'
            'step_type indicates the dm_env.StepType of that step, i.e. '
            'whether it was first, mid, or last.'
            '\n\n'
            '\n\n'
            'action contains the agent action for the step.'
            '\n\n'
            '\n\n'
            'meta_state is the serialized meta_state of the environment.'
            '\n\n'
            '\n\n'
            'state is a list, each element of which represents a layer in the '
            'environment state. The layer is represented as a list [k, [], [], '
            '[], ...], where k is the layer name and the subsequent elements '
            'are serialized sprites. Each serialized sprite is a list of '
            'attributes. See attributes.txt for the attributes contained.'
        )
        if self._log_vertices == VertexLogging.ALWAYS:
            description += (
                ' Furthermore, a list of vertices is appended to the attribute '
                'list for each serialized sprite.'
            )
        elif self._log_vertices == VertexLogging.WHEN_NECESSARY:
            description += (
                '\n\n'
                '\n\n'
                'Furthermore, a list of vertices is appended to the attribute '
                'list for a serialized for the first timestep in which that '
                'serialized sprite appears, or when the sprite has changed '
                'shape.'
            )
        with open(description_filename, 'w') as f:
            f.write(description)

    def _serialize_sprite(self, s):
        """Serialize a sprite as a list of attributes."""
        attributes = [_serialize(getattr(s, x)) for x in self._attributes]
        if (self._log_vertices == VertexLogging.ALWAYS or
            (self._log_vertices == VertexLogging.WHEN_NECESSARY and
                s.just_set_shape)):
            attributes.append(s.vertices.tolist())
        s.just_set_shape = False
        return attributes

    def _serialized_state(self):
        """Serialized a state."""
        serialized_state = [
            [k, [self._serialize_sprite(s) for s in self.state[k]]]
            for k in self.state
        ]
        return serialized_state

    def step(self, action):
        """Step the environment with an action, logging timesteps."""
        timestep = self._environment.step(action)
        str_timestep = (
            [['time', time.time()],
             ['reward', timestep.reward],
             ['step_type', timestep.step_type.value],
             ['action', _serialize(action)],
             ['meta_state', _serialize(self._environment.meta_state)],
             self._serialized_state()]
        )
        self._episode_log.append(str_timestep)
        
        if timestep.last():
            # Write the episode to a log file
            episode_count_str = str(self._episode_count).zfill(_FILENAME_ZFILL)
            filename = os.path.join(self._log_dir, episode_count_str)
            logging.info('Logging episode {} to {}.'.format(
                self._episode_count, filename))
            with open(filename, 'w') as f:
                json.dump(self._episode_log, f)
            self._episode_count += 1
            self._episode_log = []

        return timestep



"""Abstract physic class."""

import abc


class AbstractPhysics(abc.ABC):
    """Abstract physics class.
    
    All physics objects must inherit from this class.
    """

    def __init__(self, updates_per_env_step):
        self._updates_per_env_step = updates_per_env_step
    
    @abc.abstractmethod
    def apply_physics(self, state, updates_per_env_step):
        """Step the physics.
        
        Args:
            state: Environment state.
            updates_per_env_step: Int. Number of times the physics is applied
                per environment step.
        """
        pass

    def reset(self, state):
        """Reset the physics.

        This is called at every episode reset of the environment. Usually it
        does nothing (usually the physics don't change between episodes), but
        for maze environments this is used to initialize the maze for the
        episode.

        Args:
            state: Environment state.
        """
        pass

    def step(self, state):
        """Step physics, applying forces self._updates_per_env_step times."""
        for _ in range(self._updates_per_env_step):
            self.apply_physics(state, self._updates_per_env_step)

    @property
    def updates_per_env_step(self):
        """Number of times the physics is applied per environment step."""
        return self._updates_per_env_step




"""Pong task.

This task is based on the task in the following paper:
Rajalingham, Rishi and Piccato, Aida and Jazayeri, Mehrdad (2021). The role of
mental simulation in primate physical inference abilities.

In this task the subject controls a paddle at the bottom of the screen with a
joystick. The paddle is constrained to only move left-right. Each trial one ball
falls from the top of the screen, starting at a random position and moving with
a random angle. The ball bounces off of vertical walls on either side of the
screen as it falls. The subject's goal is the intercept the ball with the
paddle.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import game_rules
from moog import observers
from moog import physics as physics_lib
from moog import sprite
from moog import tasks
from moog.state_initialization import distributions as distribs


def get_config(_):
    """Get environment config."""

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Occluder
    occluder_shape = np.array([
        [-0.1, 0.2], [1.1, 0.2], [1.1, 0.6], [-0.1, 0.6]
    ])
    occluder = sprite.Sprite(
        x=0., y=0., shape=occluder_shape, scale=1., c0=0.6, c1=1., c2=1.)
    
    # Prey 
    prey_factors = distribs.Product(
        [distribs.Continuous('x', 0.1, 0.8),
         distribs.Continuous('x_vel', -0.03, 0.03)],
        y=1.2, y_vel=-0.02, shape='circle', scale=0.07, c0=0.2, c1=1., c2=1.,
    )

    # Walls
    left_wall = [[0.05, -0.2], [0.05, 2], [-1, 2], [-1, -0.2]]
    right_wall = [[0.95, -0.2], [0.95, 2], [2, 2], [2, -0.2]]
    walls = [
        sprite.Sprite(shape=np.array(v), x=0, y=0, c0=0., c1=0., c2=0.5)
        for v in [left_wall, right_wall]
    ]

    def state_initializer():
        agent = sprite.Sprite(
            x=0.5, y=0.1, shape='square', aspect_ratio=0.2, scale=0.1, c0=0.33,
            c1=1., c2=0.66)
        state = collections.OrderedDict([
            ('walls', walls),
            ('prey', [sprite.Sprite(**prey_factors.sample())]),
            ('agent', [agent]),
            ('occluder', [occluder]),
        ])
        return state

    ############################################################################
    # Physics
    ############################################################################

    agent_friction_force = physics_lib.Drag(coeff_friction=0.25)
    asymmetric_collision = physics_lib.Collision(
        elasticity=1., symmetric=False, update_angle_vel=False)
    physics = physics_lib.Physics(
        (agent_friction_force, 'agent'),
        (asymmetric_collision, 'prey', 'walls'),
        updates_per_env_step=10,
    )

    ############################################################################
    # Task
    ############################################################################

    contact_task = tasks.ContactReward(1., layers_0='agent', layers_1='prey')
    reset_task = tasks.Reset(
        condition=lambda state: all([s.y < 0. for s in state['prey']]),
        steps_after_condition=15,
    )
    task = tasks.CompositeTask(contact_task, reset_task)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Joystick(
        scaling_factor=0.005, action_layers='agent', constrained_lr=True)

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(
        image_size=(64, 64), anti_aliasing=1, color_to_rgb='hsv_to_rgb')

    ############################################################################
    # Game rules
    ############################################################################

    prey_vanish = game_rules.VanishOnContact(
        vanishing_layer='prey',
        contacting_layer='agent',
    )
    rules = (prey_vanish,)

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': rules,
    }
    return config



"""Maze walk classes.

The classes in this file can be used to walk sprites in mazes, either randomly
or with a deterministic policy. They should be used as forces in a
.physics.Physics instance.
"""

import abc
import numpy as np
from moog import physics
from moog import maze_lib

# Small float to snap positions to grid
_EPSILON = 1e-5


class AbstractMazeWalk(physics.AbstractForce, metaclass=abc.ABCMeta):
    """Abstract maze walk class.
    
    All maze walk classes should inherit from this.
    """

    def __init__(self, speed, maze_layer='walls'):
        """Constructor.

        Args:
            speed: Constant speed at which sprites walk.
            maze_layer: String. Layer in the environment state containing the
                maze wall sprites.
        """
        self._speed = speed
        self._maze_layer = maze_layer

    def reset(self, state):
        """Resetting re-infers the maze from the state."""
        self._maze = maze_lib.Maze.from_state(
            state, maze_layer=self._maze_layer)

    def step(self, *sprites, updates_per_env_step=1):
        """Step the sprites, updating their velocities."""
        for sprite in sprites:
            self._step_sprite(sprite, updates_per_env_step=updates_per_env_step)

    @abc.abstractmethod
    def _step_sprite(self, sprite, updates_per_env_step=1):
        """Step a sprite, updating its velocity."""
        pass

    def _get_pos_vel(self, sprite, updates_per_env_step=1):
        """Get position, velocity, and whether entering intersection.
        
        Args:
            sprite: Sprite instance.
            updates_per_env_step: Int. Number of physics steps per environment
                step.

        Returns:
            position: Position of the sprite.
            velocity: Velocity of the sprite, adjusted to have self._speed
                speed if nonzero.
            entering_intersection: Bool. Whether or not the sprite is entering
                an intersection in the next step.
        """
        position = sprite.position
        velocity = self._speed * np.sign(sprite.velocity)
        next_position = position + velocity / updates_per_env_step

        intersection = (
            self._maze.grid_side * self._get_nearest_point(position) +
            self._maze.half_grid_side)

        dist_next_current = sum(np.abs(next_position - position))
        dist_intersection_next = sum(np.abs(next_position - intersection))
        dist_intersection_current = sum(np.abs(next_position - intersection))

        entering_intersection = (
            dist_next_current > dist_intersection_current and 
            dist_next_current > dist_intersection_next)

        return position, velocity, entering_intersection

    def _get_nearest_point(self, position):
        """Get nearest point of the maze to a position.
        
        Args:
            position: Float array of size (2,).

        Returns:
            nearest_inds: Int array of size (2,). Indices of the nearest maze
                point to position.
        """
        nearest_inds = np.round(position / self._maze.grid_side - 0.5)
        return nearest_inds.astype(int)


class RandomMazeWalk(AbstractMazeWalk):
    """Random maze walk."""

    def __init__(self, speed, maze_layer='walls', prevent_backtracking=True,
                 allow_wall_backtracking=False, only_turn_at_wall=False):
        """Constructor.
        
        Applying this physics to sprites makes them walk with constant speed in
        a maze, taking random turns at corners and intersections.

        Args:
            speed: Float. Speed for the sprite to move at.
            maze_layer: String. Layer in the environment state containing the
                maze sprites.
            prevent_backtracking: Bool. Whether to prevent backtracking
                (changing direction to go the opposite way).
            allow_wall_backtracking: Bool. Whether to allow backtracking if the
                sprite cannot go forward. If False, the sprite will turn when it
                hits a wall but never go back the way it came.
            only_turn_at_wall: Bool. Whether to only turn when cannot continue
                straight.
        """
        super(RandomMazeWalk, self).__init__(speed, maze_layer=maze_layer)
        self._prevent_backtracking = prevent_backtracking
        self._allow_wall_backtracking = allow_wall_backtracking
        self._only_turn_at_wall = only_turn_at_wall

    def _update_valid_directions(self, valid_directions, velocity):
        """Update valid directions based on what kinds of backtracking to allow.

        This updating occurs in place.

        Args:
            valid_directions: Binary array of shape (2, 2) indicating which
                cardinal directions are available to move in.
            velocity: Current sprite velocity.
        """
        # If not preventing backtracking, all open directions are valid
        if not self._prevent_backtracking:
            return
        axis = np.argmax(np.abs(velocity))
        direction = np.sign(velocity[axis])

        # If velocity is zero, all open directions are valid
        if direction == 0:
            return
        
        # If hit a wall and allow wall backtracking, all open directions are
        # valid
        can_continue = valid_directions[axis, int(0.5 * (1 + direction))]
        if not can_continue and self._allow_wall_backtracking:
            return
        # If not hit a wall and only turn at wall, then continue
        if can_continue and self._only_turn_at_wall:
            valid_directions.fill(0)
            valid_directions[axis, int(0.5 * (1 + direction))] = 1
            return

        # If none of the above conditions are true, prevent backtracking
        valid_directions[axis, int(0.5 * (1 - direction))] = False
    
    def _step_sprite(self, sprite, updates_per_env_step=1):
        """Update a sprite's velocity.
        
        Args:
            sprite: Sprite instance.
            updates_per_env_step: Int. Number of physics steps per environment
                step.
        """
        if np.isinf(sprite.mass):
            return

        position, velocity, entering_intersection = self._get_pos_vel(
            sprite, updates_per_env_step=updates_per_env_step)
        nearest_inds = self._get_nearest_point(position)
        
        # If sprite is entering an intersection or stationary, find the valid
        # directions to move in.
        if entering_intersection:
            valid_directions = self._maze.valid_directions(
                nearest_inds[0], nearest_inds[1])
            self._update_valid_directions(valid_directions, velocity)
        elif np.all(velocity == 0.):
            rounded_position = (
                self._maze.half_grid_side + nearest_inds * self._maze.grid_side)
            on_grid = np.abs(rounded_position - position) < _EPSILON
            if np.all(on_grid):
                valid_directions = self._maze.valid_directions(
                    nearest_inds[0], nearest_inds[1])
            else:
                valid_directions = np.zeros((2, 2))
                valid_directions[1 - np.argmax(on_grid)] = 1
        else:
            sprite.velocity = velocity
            return
        
        # Sample new direction to move in
        sample = valid_directions * np.random.rand(2, 2)
        sample_ind = np.argmax(np.ravel(sample))

        # Update velocity to move in new direction, but don't eliminate current
        # velocity as that might be needed to get us to the intersection
        velocity[sample_ind // 2] = (
            (1 + _EPSILON) * self._speed * (2 * (sample_ind % 2) - 1))
        sprite.velocity = velocity


class DeterministicMazeWalk(AbstractMazeWalk):
    """Deterministic maze walk."""

    def __init__(self, speed, step_velocities, maze_layer='walls'):
        """Constructor.

        Args:
            speed: Float. Speed for the sprite to move at.
            step_velocities: Iterable of 2-iterables. Each element is the
                perscribed velocity of a sprite at a timestep. The elements are
                read out front-to back each time a sprite enters an
                intersection. Note that to handle multiple sprites you have to
                interleave their velocities, since this class has no way of
                knowing which sprite it is stepping.
            maze_layer: String. Layer in the environment state containing the
                maze sprites.
        """
        super(DeterministicMazeWalk, self).__init__(
            speed, maze_layer=maze_layer)
        self._step_velocities = [np.array(v) for v in step_velocities]

    def _step_sprite(self, sprite, updates_per_env_step=1):
        """Update a sprite's velocity.
        
        Args:
            sprite: Sprite instance.
            updates_per_env_step: Int. Number of physics steps per environment
                step.
        """
        _, velocity, entering_intersection = self._get_pos_vel(
            sprite, updates_per_env_step=updates_per_env_step)

        if entering_intersection or np.all(velocity == 0.):
            # If self._step_velocities is not empty use the next one. Otherwise,
            # do nothing.
            if len(self._step_velocities) > 0:
                new_velocity = self._step_velocities.pop(0)
                if np.any(np.sign(new_velocity) != np.sign(velocity)):
                    sprite.velocity = np.clip(
                        (1 - _EPSILON) * velocity + new_velocity, -self._speed,
                        -self._speed)



"""Discrete grid action space for controlling agent avatars."""

from . import abstract_action_space
from dm_env import specs
import numpy as np


class Grid(abstract_action_space.AbstractActionSpace):
    """Discrete grid action space.
    
    This action space has 5 actions {left, right, up, down, do-nothing}. These
    actions control either the force or the velocity of the agent(s).
    """

    _ACTIONS = (
        np.array([-1, 0]),  # Move left
        np.array([1, 0]),  # Move right
        np.array([0, -1]),  # Move down
        np.array([0, 1]),  # Move up
        np.array([0, 0]),  # Do not move
    )

    def __init__(self, scaling_factor=1., action_layers='agent',
                 control_velocity=False, momentum=0.):
        """Constructor.
        
        Args:
            scaling_factor: Scalar. Scaling factor multiplied to the action.
            agent_layer: String or iterable of strings. Elements (or itself if
                string) must be keys in the environment state. All sprites in
                these layers will be acted upon by this action space.
            control_velocity: Bool. Whether to control velocity (True) or force
                (False).
            momentum: Float in [0, 1]. Discount factor for previous action. This
                should be zero if control_velocity is False, because imparting
                forces automatically gives momentum to the agent(s) being
                controlled. If control_velocity is True, setting this greater
                than zero gives the controlled agent(s) momentum. However, the
                velocity is clipped at scaling_factor, so the agent only retains
                momentum when stopping or changing direction and does not
                accelerate.
        """
        self._scaling_factor = scaling_factor
        if not isinstance(action_layers, (list, tuple)):
            action_layers = (action_layers,)
        self._action_layers = action_layers
        self._control_velocity = control_velocity
        self._momentum = momentum

        self._action_spec = specs.DiscreteArray(len(self._ACTIONS))

    def step(self, state, action):
        """Apply action to environment state.

        Args:
            state: Ordereddict of layers of sprites. Environment state.
            action: Numpy float array of size (2). Force to apply.
        """
        self._action *= self._momentum

        self._action += Grid._ACTIONS[action]
        self._action = np.clip(
            self._action, -self._scaling_factor, self._scaling_factor)
            
        for action_layer in self._action_layers:
            for sprite in state[action_layer]:
                if self._control_velocity:
                    sprite.velocity = self._action / sprite.mass
                else:
                    sprite.velocity += self._action / sprite.mass

    def reset(self, state):
        """Reset action space at start of new episode."""
        del state
        self._action = np.zeros(2)

    def random_action(self):
        """Return randomly sampled action."""
        return np.random.randint(len(Grid._ACTIONS))
    
    def action_spec(self):
        return self._action_spec



"""Raw state observer."""

from . import abstract_observer


class RawState(abstract_observer.AbstractObserver):
  """Passes the raw state itself as observation.
  
  The raw state is an OrderedDict of iterables of sprites. Observing the raw
  state can be useful for example for hand-crafted agents with particular
  policies in multi-agent tasks.
  """

  def __init__(self):
    pass

  def __call__(self, state):
    """Passes the raw state straight through."""
    return state
  
  def observation_spec(self):
    raise NotImplementedError



"""Abstract action space."""

import abc


class AbstractActionSpace(abc.ABC):
    """Abstract action space class.
    
    All action spaces must inherit from this class.
    """
    
    @abc.abstractmethod
    def step(self, state, action):
        """Apply action to environment state.

        The action may do things like change sprites' positions or velocities,
        etc. Any change to the state of the environment can be implemented by an
        action space.
        
        Args:
            state: Environment state. OrderedDict of iterables of sprites.
            action: Action object. Type and size depends on action space.
        """
        pass

    def reset(self, state):
        """Reset action space at start of new episode.

        Action spaces are often state-less, but may be stateful (e.g. have a
        momentum attribute), so they need to be reset.

        Args:
            state: Environment state. OrderedDict of iterables of sprites.
        """
        pass

    @abc.abstractmethod
    def random_action(self):
        """Sample random action from action space."""
        pass

    @abc.abstractmethod
    def action_spec(self):
        """Get action spec for the output.
        
        Returns:
            dm_env.specs.ArraySpec or nested structure of such.
        """
        pass



"""First-person shape catcher task with parallelogram arrangement of prey.

In this task the subject controls a green circlular agent with a joystick. The
motion is first-person, so the agent is fixed at the center of the screen while
everything else moves. There are four yellow prey sprites. These prey sprites
are identical parallelograms, and they are spatially arranged in a parallelogram
with the same aspect ratio. There is an annulus occluding all peripheral vision,
i.e. never are two prey visible simultaneously.

The entire prey configuration may be drifting and rotating, depending on the
level. See the get_config() function at the bottom of this file.

This forces the subject to make a hierarchical inference task. After seeing the
first prey, there are four possible arrangements of the other prey. After
finding the second prey, there are two possible arrangements of the remaining
prey. After finding the third prey, the fourth prey's position is deterministic.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import game_rules
from moog import observers
from moog import physics as physics_lib
from moog import shapes
from moog import sprite
from moog import tasks

# Width and height of each cell in the background grid
_GRID_SIZE = 0.3


def get_parallelogram(min_axis_ratio=0.4):
    """Get parallelogram vertices centered around 0 with maximum radius 1."""
    angles = np.pi * (np.array([0., 0.5, 1., 1.5]) + np.random.uniform(0, 2))
    vertices = np.stack((np.sin(angles), np.cos(angles)), axis=1)
    axis_ratio = np.random.uniform(min_axis_ratio, 1.)
    vertices *= np.array([[1.], [axis_ratio], [1.], [axis_ratio]])

    return vertices


def get_prey(centered_vertices, scale=1., max_vel=0., sprite_scale=0.1):
    """Get prey sprites.

    Args:
        centered_vertices: Numpy array of shape [num_vertices, 2] containing
            vertex positions.
        scale: Re-scaling factor of centered_vertices for the global space.
        max_vel: Maximum velocity of the sprites.
        sprite_scale: Re-scaling factor of centered_vertices for the individual
            sprite shapes.
    """
    sprite_shape = sprite_scale * centered_vertices
    sprite_positions = scale * centered_vertices
    sprite_positions += np.array([0.5, 0.5]) - sprite_positions[0]

    # We sample each sprite's velocity independently so that the entire tethered
    # configuration may rotate
    prey = [
        sprite.Sprite(
            x=pos[0], y=pos[1], shape=sprite_shape, scale=1.,
            x_vel=np.random.uniform(-1 * max_vel, max_vel),
            y_vel=np.random.uniform(-1 * max_vel, max_vel),
            c0=0.2, c1=1., c2=1.)
        for pos in sprite_positions
    ]
    return prey


def _get_config(max_vel):
    """Get environment config."""

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Grid
    grid = shapes.grid_lines(
        grid_x=_GRID_SIZE, grid_y=_GRID_SIZE, buffer_border=1., c0=0., c1=0.,
        c2=0.5)

    def state_initializer():
        agent = sprite.Sprite(
            x=0.5, y=0.5, shape='circle', scale=0.04, c0=0.33, c1=1., c2=0.66)
        annulus_shape = shapes.annulus_vertices(0.15, 2.)
        agent_annulus = sprite.Sprite(
            x=0.5, y=0.5, shape=annulus_shape, scale=1., c0=0.6, c1=1., c2=1.)
        prey = get_prey(
            get_parallelogram(min_axis_ratio=0.5),
            scale=0.4,
            max_vel=max_vel,
            sprite_scale=0.075,
        )
        state = collections.OrderedDict([
            ('grid', grid),
            ('prey', prey),
            ('agent', [agent]),
            ('agent_annulus', [agent_annulus]),
        ])
        return state

    ############################################################################
    # Physics
    ############################################################################

    force = (physics_lib.Drag(coeff_friction=0.25), ['agent', 'agent_annulus'])
    corrective_physics = physics_lib.Tether(('prey',), update_angle_vel=True)
    physics = physics_lib.Physics(
        force,
        updates_per_env_step=10,
        corrective_physics=corrective_physics,
    )

    ############################################################################
    # Task
    ############################################################################

    prey_task = tasks.ContactReward(
        1, layers_0='agent', layers_1='prey',
        condition=lambda s_agent, s_prey: s_prey.c1 > 0.5,
    )
    reset_trial_task = tasks.Reset(
        condition=lambda state: all([s.c1 < 0.5 for s in state['prey']]),
        steps_after_condition=10,
    )
    task = tasks.CompositeTask(prey_task, reset_trial_task, timeout_steps=500)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Joystick(
        scaling_factor=0.01, action_layers=('agent', 'agent_annulus'))

    ############################################################################
    # Observer
    ############################################################################

    _polygon_modifier = observers.polygon_modifiers.FirstPersonAgent(
        agent_layer='agent')
    observer = observers.PILRenderer(
        image_size=(64, 64),
        anti_aliasing=1,
        color_to_rgb='hsv_to_rgb',
        polygon_modifier=_polygon_modifier,
    )

    ############################################################################
    # Game rules
    ############################################################################

    # Make prey gray upon contact
    def _make_prey_gray(prey):
        prey.c1 = 0.
        prey.c2 = 0.6
    make_prey_gray = game_rules.ModifyOnContact(
        layers_0='agent',
        layers_1='prey',
        modifier_1=_make_prey_gray,
    )

    # Keep agent near center
    keep_near_center = game_rules.KeepNearCenter(
        agent_layer='agent',
        layers_to_center=['agent_annulus', 'prey'],
        grid_x=_GRID_SIZE,
    )

    rules = (make_prey_gray, keep_near_center)

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': rules,
    }
    return config


def get_config(level):
    """Get config dictionary of kwargs for environment constructor.
    
    Args:
        level: Int. Different values yield different velocities of the prey.
    """
    if level == 0:
        return _get_config(max_vel=0.)
    elif level == 1:
        return _get_config(max_vel=0.01)
    elif level == 2:
        return _get_config(max_vel=0.02)
    else:
        raise ValueError('Invalid level {}'.format(level))



# This file was forked and heavily modified from the file here:
# https://github.com/deepmind/spriteworld/blob/master/spriteworld/sprite.py
# Here is the license header for that file:

# Copyright 2019 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Sprite.

The main class in this file is Sprite. Instances of Sprite comprise the state of
an environment. This file also contains some helper functions for updating the 
attributes of a sprite and computing crossing points of lines and sprite edges.
"""

import collections
from matplotlib import path as mpl_path
from matplotlib import transforms as mpl_transforms
import numpy as np
from moog import shapes

GLOBAL_SPRITE_COUNT = 0

# Tiny float for numerical stability in segment_crossings()
_EPSILON_INTERPOLATION = 1e-8


def _cross_2d(x, y):
    """2-dimensional cross product.
    
    Args:
        x: Numpy array of shape [..., 2]. First vector.
        y: Numpy array of shape [..., 2]. Second vector.
        
    Returns:
        cross: Numpy array of shape [...]. Cross product of x and y.
    """
    return x[..., 0] * y[..., 1] - x[..., 1] * y[..., 0]


def update_sprite(sprite, **factors):
    """Update sprite in place given an entirely new set of factors.

    This file updates a sprite as efficienctly as possible, i.e. without
    resetting the shape or transforming the path unless those are necessary.

    Args:
        sprite: Instance of Sprite.
        **factors: Dict. Keys must be in Sprite.FACTOR_NAMES. Values are new
            values for those factors.
    """
    # Some factors can be set directly
    for k in ['c0', 'c1', 'c2', 'opacity', 'angle_vel', 'mass', 'metadata']:
        if k in factors:
            setattr(sprite, k, factors[k])

    # Set position
    if 'x' in factors or 'y' in factors:
        sprite.position = [
            factors.get('x', sprite.position[0]),
            factors.get('y', sprite.position[1]),
        ]

    # Set velocity
    if 'x_vel' in factors or 'y_vel' in factors:
        sprite.velocity = [
            factors.get('x_vel', sprite.velocity[0]),
            factors.get('y_vel', sprite.velocity[1]),
        ]

    # Setting shape, angle, scale, or aspect_ratio requires transforming the
    # sprite path, which is somewhat computationally expensive, so for these we
    # check to see if they've changed before setting them.
    for k in ['angle', 'scale', 'aspect_ratio']:
        if k in factors and getattr(sprite, k) != factors[k]:
            setattr(sprite, k, factors[k])
    if 'shape' in factors:
        if isinstance(factors['shape'], str):
            # Shape is encoded as a string, not as vertices, so we can just
            # reset the shape attribute
            if sprite.shape != factors['shape']:
                sprite.shape = factors['shape']
        else:
            # Shape is encoded as an array of vertices.
            if not np.array_equal(sprite.vertices, factors['shape']):
                # Note that we could invert the angle, scale, aspect_ratio, and
                # position transformations to then use the sprite's shape
                # setter, but this would add about 15 lines of code and would
                # not be very readable. So instead we take a shortcut and
                # directly override the protected sprite._path attribute.
                vertices = np.concatenate(
                    (factors['shape'], factors['shape'][:1]), axis=0)
                sprite._path = mpl_path.Path(vertices)  #pylint: disable=protected-access

    return


def segment_crossing_coefficients(start_0, end_0, start_1, end_1):
    """Solves linear equations to compute crossing points of segments.

    Finding the crossing point between segment [start_0, end_0] and segment
    [start_1, end_1] can be done by solving
        start_0 + A * delta_0 = start_1 + B * delta_1
    for A and B, where delta_i = end_i - start_i.

    The solution (by a little vector algebra) is
        A = ((start_1 - start_0) x delta_1) / (delta_0 x delta_1)
        B = ((start_1 - start_0) x delta_0) / (delta_0 x delta_1)
    where x denotes cross product. If A and B are both in [0, 1], then the
    segments [start_0, end_0] and [start_1, end_1] intersect at point
    start_0 + A * delta_0.

    This function does this computation vectorized to handle multiple segments
    at once, returning the coefficients A and B. Namely, it computes all
    pairwise crossings between two arrays of segments.

    The reason this function is factored out of segment_crossings() below is
    because some calling code (e.g. collisions) need to use the A and B
    coefficients.

    Args:
        start_0: Numpy array of shape [N, 2]. Starting points for set 0.
        end_0: Numpy array of shape [N, 2]. Ending points for set 0.
        start_1: Numpy array of shape [M, 2]. Starting points for set 1.
        end_1: Numpy array of shape [M, 2]. Ending points for set 1.

    Returns:
        A: Numpy array of shape [N, N]. A[i, j] is the coefficient A (see above)
            for the crossing of segment [start_0[i], end_0[i]] and the segment
            [start_1[j], end_1[j]].
        B: Numpy array of shape [N, N]. B[i, j] is the coefficient B (see above)
            for the crossing of segment [start_0[i], end_0[i]] and the segment
            [start_1[j], end_1[j]].
    """
    s_0 = start_0
    ds_0 = end_0 - start_0
    s_1 = start_1
    ds_1 = end_1 - start_1

    # Expand the _0's and _1's in different dimensions to leverage broadcasting.
    s_0 = s_0[:, np.newaxis]
    ds_0 = ds_0[:, np.newaxis]
    s_1 = s_1[np.newaxis]
    ds_1 = ds_1[np.newaxis]

    # Need to add small epsilon for stability, else may divide by zero below
    ds_0_cross_ds_1 = _cross_2d(ds_0, ds_1) + _EPSILON_INTERPOLATION
    s_1_minus_s_0 = s_1 - s_0

    A = _cross_2d(s_1_minus_s_0, ds_1) / ds_0_cross_ds_1
    B = _cross_2d(s_1_minus_s_0, ds_0) / ds_0_cross_ds_1

    return A, B


def segment_crossings(start_0, end_0, start_1, end_1):
    """Finds all pairwise crossings between two arrays of segments.

    See segment_crossing_coefficients() documentation above for an explanation
    of how this is done. In contrast to segment_crossing_coefficients(), this
    function outputs the actual crossing points themselves.

    Args:
        start_0: Numpy array of shape [N, 2]. Starting points for set 0.
        end_0: Numpy array of shape [N, 2]. Ending points for set 0.
        start_1: Numpy array of shape [M, 2]. Starting points for set 1.
        end_1: Numpy array of shape [M, 2]. Ending points for set 1.

    Returns:
        crossing_points: Numpy array of shape [K, 2] containing all crossing
            points. Here K is the number of crossing points.
        inds_crossings: Numpy array of shape [K, 2] containing indices for
            segments_0 and segments_1 respectively. Specifically,
            inds_crossings[i][0] is the index in [0, N] of the set 0 segment and
            inds_crossings[i][1] is the index in [0, M] of the set 1 segment in
            crossing point i.
    """
    A, B = segment_crossing_coefficients(start_0, end_0, start_1, end_1)

    # Crossings occur when A and B are both in [0, 1]
    crossings = (A > 0) * (A < 1) * (B > 0) * (B < 1)
    inds_crossings = np.argwhere(crossings)

    # Linear combination of segment 0 with A coefficients gives crossing points
    crossing_points = np.array([
        start_0[ind_0] + A[ind_0, ind_1] * (end_0[ind_0] - start_0[ind_0])
        for (ind_0, ind_1) in inds_crossings
    ])
    return crossing_points, inds_crossings


def sprite_edge_crossings(sprite_0, sprite_1):
    """Find all points where the boundary of sprite_0 and sprite_1 cross.

    See segment_crossings() documentation for details.

    Args:
        sprite_0: Instance of Sprite.
        sprite_1: Instance of Sprite.

    Returns:
        crossing_points: Numpy float array of shape [K, 2] where K is the number
            of crossings of the boundaries of sprite_0 and sprite_1.
        inds_crossings: Numpy ind array of shape [K, 2].
    """
    path_0 = sprite_0.path.vertices
    path_1 = sprite_1.path.vertices

    crossing_points, inds_crossings = segment_crossings(
        start_0=path_0[:-1],
        end_0=path_0[1:],
        start_1=path_1[:-1],
        end_1=path_1[1:],
    )

    return crossing_points, inds_crossings


class Sprite(object):
    """Sprite class.

    Sprites are polygons parameterized by a few factors (such as position,
    shape, angle, scale, color, velocity, and mass). They are the building
    blocks of the environment and the objects upon which physics can act.
    """

    FACTOR_NAMES = (
        'x',  # x-position of sprite center-of-mass (float)
        'y',  # y-position of sprite center-of-mass (float)
        'shape',  # shape (string)
        'angle',  # angle in radians (float)
        'scale',  # size of sprite (float)
        'aspect_ratio',  # aspect ratio of sprite (float)
        'c0',  # first color component (scalar)
        'c1',  # second color component (scalar)
        'c2',  # third color component (scalar)
        'opacity',  # opacity of sprite in [0, 255]
        'x_vel',  # x-component of velocity (float)
        'y_vel',  # y-component of velocity (float)
        'angle_vel',  # angular velocity in radians/time
        'mass',  # mass
        'metadata',  # optional metadata
    )

    # Shape factor name for shapes not in shapes.SHAPES
    _CUSTOM_SHAPE = 'custom'

    # Name for a circle. Must match that in shapes.SHAPES
    _CIRCLE_NAME = 'circle'

    def __init__(self,
                 x=0.5,
                 y=0.5,
                 shape='square',
                 angle=0.,
                 scale=1.,
                 aspect_ratio=1.,
                 c0=0,
                 c1=0,
                 c2=0,
                 opacity=255,
                 x_vel=0.,
                 y_vel=0.,
                 angle_vel=0.,
                 mass=1.,
                 metadata=None):
        """Construct sprite.

        This class is agnostic to the color scheme, namely (c1, c2, c3) could be
        in RGB coordinates or HSV, HSL, etc. without this class knowing. The
        color scheme conversion for rendering must be done in the renderer.

        Args:
            x: Float in [0, 1]. x-position.
            y: Float in [0, 1]. y-position.
            shape: String or numpy array of vertices. Shape of the sprite. If
                string, must be a key of shapes.SHAPES. If array, must have
                shape [N, 2] defining the vertices of the polygonal shape.
            angle: Float. Angle in radians.
            scale: Float. Scale (size) of the sprite. This is multiplied to the
                shape vertex array when constructing the sprite, hence the
                sprite width scales linearly with respect to scape, and sprite
                area scales with power 2.
            aspect_ratio: Scalar. Height/width aspect ratio.
            c0: Scalar. First coordinate of color.
            c1: Scalar. Second coordinate of color.
            c2: Scalar. Third coordinate of color.
            opacity: Integer in [0, 255]. Opacity of sprite.
            x_vel: Float. x-velocity.
            y_vel: Float. y-velocity.
            angle_vel: Float. Angular velocity in radians/time.
            mass: Float. Mass.
            metadata: Any type. Optional metadata. If None, defaults to empty
                dictionary.
        """
        # The angle, scale, and aspect ratio must be converted to floats.
        # Sometimes the state initializer distribution will feed them in as
        # numpy scalars but this causes the updating to not work.
        self._position = np.array([x, y])
        self._angle = float(angle)
        self._scale = float(scale)
        self._aspect_ratio = float(aspect_ratio)
        self._color = (c0, c1, c2)
        self._opacity = opacity
        self._velocity = np.array([x_vel, y_vel])
        self._angle_vel = angle_vel
        self._mass = mass
        self.metadata = metadata

        # This calls shape.setter, which does shape path setting
        self.shape = shape

        # Increment the global sprite count and set self._id, which can be used
        # to identify sprite instances.
        global GLOBAL_SPRITE_COUNT
        self._id = GLOBAL_SPRITE_COUNT
        GLOBAL_SPRITE_COUNT += 1

    def _set_shape_path(self, shape_path):
        """Set shape path and moment of inertia.

        The self._shape_path set by this function is centered around the shape's
        center of mass, assuming a uniform mass distribution. Moment of inertia
        is calculated in tandem, to avoid having to iterate through vertices
        multiple times.

        The rotational moment of intertia of a 2-dimensional body about the
        z-axis (i.e. perpendicular to the plane and passing through the origin):
            I = density * integral_{area}[x^2 + y^2 dx dy]
              = I_x + I_y

        This is important to have because it is needed to relate torque to
        rotational acceleration via Newton's Law:
            torque = I * rotational_acceleration
        This is needed in our environment to simulate realistic physical
        collisions when sprites are allowed to rotate.

        In this function we compute I_x and I_y themselves and set
            self._x_y_rotational_inertia = [I_x, I_y]
        The reason to have these components is to allow them to be adjusted upon
        changes in aspect ratio without having to re-compute them from scratch
        by re-running this function. See self._set_path().

        Also see self.moment_of_inertia for the summed moment of inertia I.

        Computing I_x and I_y of a polygon can be done by breaking the polygon
        into triangles. Then for a triangle with one vertex at the origin the
        above integral can be solved analytically.
        """
        # Compute centroid, intertia, and area
        num_vertices = shape_path.shape[0]
        inertia = np.array([0., 0.])
        area = 0.
        centroid = np.array([0., 0.])
        for i in range(num_vertices):
            vertex_0 = shape_path[i]
            vertex_1 = shape_path[(i + 1) % num_vertices]

            # x and y moments of inertia
            cross = _cross_2d(vertex_0, vertex_1)
            inertia += (1. / 12.) * cross * (
                vertex_0 * vertex_0 + vertex_1 * vertex_1 + vertex_0 * vertex_1)

            # area and centroid
            triangle_area = cross / 2.
            triangle_centroid = (vertex_0 + vertex_1) / 3.
            area += triangle_area
            centroid += triangle_centroid * triangle_area
        centroid /= area

        if area < 0:
            # Path was defined clockwise, which will mess up collisions, so we
            # reverse the path and correct for the negative area and inertia.
            # The centroid is fine, because was divided by the area.
            shape_path = shape_path[::-1]
            inertia *= -1.
            area *= -1.

        # Set shape path with center of mass at origin
        center_translate = mpl_transforms.Affine2D().translate(*(-1 * centroid))
        # Make shape path be a full loop + 1 (i.e. last point in the array is
        # the first point). This makes calculating sprite overlaps easier,
        # because mpl_path.Path.intersects_path requires the full loop + 1.
        shape_path = np.concatenate((shape_path, [shape_path[0]]))
        self._shape_path = center_translate.transform_path(
            mpl_path.Path(shape_path))

        # Use parallel axis theorem to compute moment of inertia around center
        # of mass, so we don't have to iterate through the points again
        inertia -= area * np.square(centroid)
        self._x_y_rotational_inertia = inertia / area

        # We must call self._set_path() before updating self.position, becuase
        # the position setter uses self._path, which is set in self._set_path()
        self._set_path()
        self.position = self._position + centroid

        # Make sure self._just_set_shape is True
        self._just_set_shape = True

    def _set_path(self):
        """Rotate and scale self._shape path."""
        x_y_scale = np.array([self._scale, self._scale * self._aspect_ratio])
        transform = (
            mpl_transforms.Affine2D().scale(*x_y_scale) +
            mpl_transforms.Affine2D().rotate(self._angle) +
            mpl_transforms.Affine2D().translate(*self._position))
        self._path = transform.transform_path(self._shape_path)
        self._max_radius = np.max(
            np.linalg.norm(self.vertices - self._position, axis=1))

        # Adjust rotational inertia to accomodate the change in aspect ratio and
        # scale
        self._x_y_rotational_inertia *= np.square(x_y_scale)

    def update_pos_from_vel(self, delta_t):
        """Update position based on velocity."""
        self.position = self.position + delta_t * self.velocity
        if self._angle_vel:
            self.angle = self.angle + delta_t * self._angle_vel

    def contains_point(self, point):
        """Check if the point is contained in the Sprite."""
        if self.is_symmetric_circle:
            contains_point = (
                np.linalg.norm(point - self.position) < self.max_radius)
        else:
            contains_point = self.path.contains_point(point)

        return contains_point

    def contains_points(self, points):
        """Check if the points are contained in the Sprite.

        Args:
            points: Numpy array of size (N, 2) containing coordinates of N
                points.

        Returns:
            contains_points: Boolean numpy array of size (N,), indicating for
                each point whether it is in this sprite.
        """
        if self.is_symmetric_circle:
            contains_points = (
                np.linalg.norm(points - self.position, axis=1) <=
                self.max_radius)
        else:
            contains_points = self._path.contains_points(points)

        return contains_points

    def overlaps_sprite(self, sprite):
        """Check if this and the argument sprite overlap."""
        center_dist = np.linalg.norm(self.position - sprite.position)
        if center_dist > self.max_radius + sprite.max_radius:
            return False

        # WARNING: You might think that the case of two circles depends only on
        # their radius, but using this shortcut causes instability in
        # collisions, since there is currently no special treatment for circles
        # in collisions (collisions depend on the vertices themselves). So we
        # give no special treatment to circles here.

        # Note: mpl_path.Path.intersects_path() treats its input paths as
        # paths with endpoints. Namely, an input array of N vertices
        # corresponds to a path with N - 1 edges. Thus to apply it to detect
        # intersection of sprites, we must feed it an array of length
        # len(self.vertices) + 1. In fact self._path is an array of length
        # len(self.vertices) + 1, by construction (see self._set_path()).
        # Profiling note: Using mpl_path.Path.intersects_path() is slightly
        # faster than `len(sprite_edge_crossings(self, sprite)[1] > 0)`.
        overlap = mpl_path.Path.intersects_path(
            self.path, sprite.path, filled=True)
        return overlap

    @property
    def vertices(self):
        """Numpy array of vertices of the shape."""
        return self.path.vertices[:-1]

    @property
    def path(self):
        """Numpy array of length len(self.vertices) + 1, loop of the shape."""
        return self._path

    @property
    def max_radius(self):
        return self._max_radius

    @property
    def is_symmetric_circle(self):
        return self.shape == Sprite._CIRCLE_NAME and self.aspect_ratio == 1

    @property
    def x(self):
        return self._position[0]

    @property
    def y(self):
        return self._position[1]

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, shape):
        if isinstance(shape, str) and shape in shapes.SHAPES:
            self._shape = shape
            shape_path = shapes.SHAPES[shape]
        else:
            self._shape = Sprite._CUSTOM_SHAPE
            shape_path = shape

        self._set_shape_path(shape_path)

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, a):
        if id(a) == id(self._angle):
            # See comment in @position.setter for why we catch this.
            raise ValueError(
                'Cannot call in-place operations on sprite.angle.')
        rotate = mpl_transforms.Affine2D().rotate_around(
            self.x, self.y, a - self._angle)
        self._path = rotate.transform_path(self._path)
        self._angle = a

    @property
    def scale(self):
        return self._scale

    @scale.setter
    def scale(self, s):
        self._scale = s
        self._set_path()

    @property
    def aspect_ratio(self):
        return self._aspect_ratio

    @aspect_ratio.setter
    def aspect_ratio(self, a):
        self._aspect_ratio = a
        self._set_path()

    @property
    def c0(self):
        return self._color[0]

    @c0.setter
    def c0(self, c0):
        self._color = (c0, self.c1, self.c2)

    @property
    def c1(self):
        return self._color[1]

    @c1.setter
    def c1(self, c1):
        self._color = (self.c0, c1, self.c2)

    @property
    def c2(self):
        return self._color[2]

    @c2.setter
    def c2(self, c2):
        self._color = (self.c0, self.c1, c2)

    @property
    def opacity(self):
        return self._opacity

    @opacity.setter
    def opacity(self, opacity):
        self._opacity = opacity

    @property
    def x_vel(self):
        return self._velocity[0]

    @property
    def y_vel(self):
        return self._velocity[1]

    @property
    def mass(self):
        return self._mass

    @mass.setter
    def mass(self, mass):
        self._mass = mass

    @property
    def color(self):
        return self._color

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, pos):
        if id(pos) == id(self.position):
            # This setter references the pre-set position self._position, but
            # with an in-place operation, the pre-set position is the same array
            # as `pos`, the position to be set, so this setter function does not
            # work as intended. In practice, this means that an in-place
            # operation updates self._position but doesn't update self._path,
            # which causes sprites to not move. Consequently, calling code must
            # not update position in place. That can cause devious bugs if
            # undetected, so we catch it with this ValueError.
            raise ValueError(
                'Cannot call in-place operations on sprite.position.')
        if not isinstance(pos, np.ndarray):
            pos = np.array(pos)
        translate = mpl_transforms.Affine2D().translate(*pos - self._position)
        self._path = translate.transform_path(self._path)
        self._position = pos

    @property
    def velocity(self):
        return self._velocity

    @velocity.setter
    def velocity(self, vel):
        if not isinstance(vel, np.ndarray):
            vel = np.array(vel)
        self._velocity = vel

    @property
    def angle_vel(self):
        return self._angle_vel

    @angle_vel.setter
    def angle_vel(self, angle_vel):
        self._angle_vel = angle_vel

    @property
    def just_set_shape(self):
        """This property can used and set by loggers."""
        return self._just_set_shape

    @just_set_shape.setter
    def just_set_shape(self, just_set_shape):
        self._just_set_shape = just_set_shape

    @property
    def moment_of_inertia(self):
        return sum(self.mass * self._x_y_rotational_inertia)

    @property
    def id(self):
        return self._id

    @property
    def factors(self):
        factors = collections.OrderedDict()
        for factor_name in Sprite.FACTOR_NAMES:
            factors[factor_name] = getattr(self, factor_name)
        return factors



"""Demo script.

This script can be used to play with and test prototype tasks with a matplotlib
interface. This demo does not run with a fast frame-rate or control timing well,
so is not intended for colleting high-fidelity behavior (use a psychophysics
toolbox like MWorks for that), but is intended instead to sanity-check task
prototypes.

Run with the following:
$ python3 run_demo.py --config=path.to.your.config

See also the flags at the top of this file. There are options for rendering
settings, recording to gif, and logging the behavior.

To exit the demo on a mac, press the 'esc' key. To customize key bindings, see
the key bindings in human_agent.py. If you are playing a task with an action
space that is not supported by the interfaces in gui_frames.py, please add a
custom gui interface.
"""

import sys
sys.path.insert(0, '..')

from absl import app
from absl import flags
import importlib
import os

from moog import env_wrappers
from moog import observers
from moog import environment
from moog_demos import gif_writer as gif_writer_lib
from moog_demos import human_agent

FLAGS = flags.FLAGS
flags.DEFINE_string('config',
                    'moog_demos.example_configs.pong',
                    'Filename of task config to use.')
flags.DEFINE_integer('level', 0, 'Level of task config to run.')
flags.DEFINE_integer('render_size', 512,
                     'Height and width of the output image.')
flags.DEFINE_integer('anti_aliasing', 1, 'Renderer anti-aliasing factor.')
flags.DEFINE_integer('fps', 10,
                     'Upper bound on frames per second. Note: this is not an '
                     'accurate fps for the demo, since matplotlib and tkinter '
                     'introduce additional lag.')
flags.DEFINE_integer('reward_history', 30,
                     'Number of historical reward timesteps to plot.')

# Flags for gif writing
flags.DEFINE_boolean('write_gif', False, 'Whether to write a gif.')
flags.DEFINE_string('gif_file',
                    '/logs/gifs/test.gif',
                    'File path to write the gif to.')
flags.DEFINE_integer('gif_fps', 15, 'Frames per second for the gif.')

# Flags for logging timestep data
flags.DEFINE_boolean('log_data', False, 'Whether to log timestep data.')


def main(_):
    """Run interactive task demo."""
    config = importlib.import_module(FLAGS.config)
    config = config.get_config(FLAGS.level)

    config['observers']['image'] = observers.PILRenderer(
        image_size=(FLAGS.render_size, FLAGS.render_size),
        anti_aliasing=FLAGS.anti_aliasing,
        color_to_rgb=config['observers']['image'].color_to_rgb,
        polygon_modifier=config['observers']['image'].polygon_modifier,
    )

    if 'agents' in config:
        # Multi-agent demo
        agents = config.pop('agents')
        agent_name = config.pop('agent_name')
        multi_env = environment.Environment(**config)
        env = env_wrappers.MultiAgentEnvironment(
            environment=multi_env, agent_name=agent_name, **agents)
    else:
        # Single-agent demo
        env = environment.Environment(**config)

    if FLAGS.log_data:
        log_dir = os.path.join(
            'logs', FLAGS.config.split('.')[-1], str(FLAGS.level))
        env = env_wrappers.LoggingEnvironment(env, log_dir=log_dir)

    if FLAGS.write_gif:
        gif_writer = gif_writer_lib.GifWriter(
            gif_file=FLAGS.gif_file,
            fps=FLAGS.gif_fps,
        )
    else:
        gif_writer = None

    # Constructing the agent automatically starts the environment
    human_agent.HumanAgent(
        env,
        render_size=FLAGS.render_size,
        fps=FLAGS.fps,
        reward_history=FLAGS.reward_history,
        gif_writer=gif_writer,
    )


if __name__ == "__main__":
    app.run(main)



"""First-person predator-prey task in an infinite plane.

In an inifite plane, predators and prey drift around and the agent is rewarded 
for catching prey and punished for being caught by predators. The field of view
travels with the agent, keeping the agent centered at all times (first-person),
and there is an occluding annulus around the agent.

The subject receives positive reward proportional to a prey's size when a prey
is caught, and negative reward proportional to the predator's size when the
agent is caught by a predator.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import observers
from moog import physics as physics_lib
from moog import shapes
from moog import sprite
from moog import game_rules
from moog import tasks
from moog.state_initialization import distributions as distribs
from moog.state_initialization import sprite_generators

# Extent around the [0, 1] range that may be visible. Also used as distance from
# which predators and prey are spawned.
_FIELD_BUFFER = 0.7

# Distance beyond which predators vanish. Must be at least _FIELD_BUFFER.
_VANISH_DIST = 1.2

# Width and height of each cell in the background grid
_GRID_SIZE = 0.4


def _get_boundary_pos_distribution(boundary_buffer):
    """Get distribution generating positions from a square boundary."""
    boundary_range = [-1. * boundary_buffer, 1. + boundary_buffer]
    initialize_left = distribs.Product(
        [distribs.Continuous('y', *boundary_range)], x=boundary_range[0])
    initialize_right = distribs.Product(
        [distribs.Continuous('y', *boundary_range)], x=boundary_range[1])
    initialize_bottom = distribs.Product(
        [distribs.Continuous('x', *boundary_range)], y=boundary_range[0])
    initialize_top = distribs.Product(
        [distribs.Continuous('x', *boundary_range)], y=boundary_range[1])
    position_distribution = distribs.Mixture([
        initialize_left, initialize_right, initialize_bottom, initialize_top])
    return position_distribution


def _get_vel_distribution(min_vel, max_vel):
    """Get distribution generating velocities from a square annulus."""
    vel_distribution = distribs.SetMinus(
        distribs.Product([
            distribs.Continuous('x_vel', -1. * max_vel, max_vel),
            distribs.Continuous('y_vel', -1. * max_vel, max_vel),
        ]),
        hold_out=distribs.Product([
            distribs.Continuous('x_vel', -1. * min_vel, min_vel),
            distribs.Continuous('y_vel', -1. * min_vel, min_vel),
        ]),
    )
    return vel_distribution


def get_config(_):
    """Get environment config."""

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Agent
    agent = sprite.Sprite(
        x=0.5, y=0.5, shape='circle', scale=0.04, c0=0.33, c1=1., c2=0.66)
    annulus_vertices = shapes.annulus_vertices(
        inner_radius=0.08, outer_radius=0.3)
    agent_annulus = sprite.Sprite(
        x=0.5, y=0.5, shape=annulus_vertices, scale=1., c0=0.6, c1=1., c2=1.)
    
    # Predator generator
    max_predator_vel = 0.02
    predator_pos = _get_boundary_pos_distribution(_FIELD_BUFFER)
    predator_vel = _get_vel_distribution(
        0.5 * max_predator_vel, max_predator_vel)
    predator_factors = distribs.Product(
        [predator_pos, predator_vel,
         distribs.Continuous('scale', 0.07, 0.13)],
        shape='circle', c0=0., c1=1., c2=0.8,
    )

    # Prey generator
    max_prey_vel = 0.01
    prey_pos = _get_boundary_pos_distribution(_FIELD_BUFFER)
    prey_vel = _get_vel_distribution(0.5 * max_prey_vel, max_prey_vel)
    prey_factors = distribs.Product(
        [prey_pos, prey_vel,
         distribs.Continuous('scale', 0.07, 0.13)],
        shape='circle', c0=0.2, c1=1., c2=1.,
    )

    # Grid
    grid = shapes.grid_lines(
        grid_x=_GRID_SIZE, grid_y=_GRID_SIZE, buffer_border=1., c0=0., c1=0.,
        c2=0.5)

    def state_initializer():
        state = collections.OrderedDict([
            ('grid', grid),
            ('prey', []),
            ('agent', [agent]),
            ('predators', []),
            ('agent_annulus', [agent_annulus]),
        ])
        return state

    ############################################################################
    # Physics
    ############################################################################

    agent_friction_force = physics_lib.Drag(coeff_friction=0.25)
    physics = physics_lib.Physics(
        (agent_friction_force, ['agent', 'agent_annulus']),
        updates_per_env_step=10,
    )

    ############################################################################
    # Task
    ############################################################################

    def _predator_reward_fn(_, predator_sprite):
        return -2. * predator_sprite.scale
    predator_task = tasks.ContactReward(
        reward_fn=_predator_reward_fn,
        layers_0='agent',
        layers_1='predators',
        reset_steps_after_contact=0,
    )
    def _prey_reward_fn(_, prey_sprite):
        return prey_sprite.scale
    prey_task = tasks.ContactReward(
        reward_fn=_prey_reward_fn,
        layers_0='agent',
        layers_1='prey',
    )
    task = tasks.CompositeTask(predator_task, prey_task)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Joystick(
        scaling_factor=0.003,
        action_layers=('agent', 'agent_annulus'),
        constrained_lr=False,
    )

    ############################################################################
    # Observer
    ############################################################################

    _polygon_modifier = observers.polygon_modifiers.FirstPersonAgent(
        agent_layer='agent')
    observer = observers.PILRenderer(
        image_size=(64, 64),
        anti_aliasing=1,
        color_to_rgb='hsv_to_rgb',
        polygon_modifier=_polygon_modifier,
    )

    ############################################################################
    # Game rules
    ############################################################################

    # Make predators appear randomly
    predator_appear_generator = sprite_generators.generate_sprites(
        predator_factors, num_sprites=1)
    predator_appear = game_rules.ConditionalRule(
        condition=lambda state: np.random.binomial(1, p=0.5),
        rules=game_rules.CreateSprites('predators', predator_appear_generator),
    )

    # Make prey appear randomly
    prey_appear_generator = sprite_generators.generate_sprites(
        prey_factors, num_sprites=1)
    prey_appear = game_rules.ConditionalRule(
        condition=lambda state: np.random.binomial(1, p=0.2),
        rules=game_rules.CreateSprites('prey', prey_appear_generator),
    )

    # Make predators and prey vanish when they are distant enough and moving
    # away.
    vanish_range = [-1. * _VANISH_DIST, 1. + _VANISH_DIST]
    def _should_vanish(s):
        pos_too_small = (s.position < vanish_range[0]) * (s.velocity < 0.)
        pos_too_large = (s.position > vanish_range[1]) * (s.velocity > 0.)
        return any(pos_too_small) or any(pos_too_large)
    predator_vanish = game_rules.VanishByFilter('predators', _should_vanish)
    prey_vanish = game_rules.VanishByFilter('prey', _should_vanish)

    # Keep agent near center
    keep_near_center = game_rules.KeepNearCenter(
        agent_layer='agent',
        layers_to_center=['agent_annulus', 'predators', 'prey'],
        grid_x=_GRID_SIZE,
    )

    # Make prey vanish when contacted by agent
    prey_caught = game_rules.VanishOnContact(
        vanishing_layer='prey', contacting_layer='agent')

    rules = (predator_appear,
             prey_appear,
             prey_vanish,
             predator_vanish,
             keep_near_center,
             prey_caught)

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': rules,
    }
    return config



"""Rules that modify environment meta_state."""

from . import abstract_rule
import itertools
import numpy as np


class ModifyMetaState(abstract_rule.AbstractRule):
    """Modify meta_state.
    
    This can be used for any in-place meta_state modification.
    """

    def __init__(self, modifier):
        """Constructor.

        Args:
            modifier: Function taking in meta_state and modifying it in place.
        """
        self._modifier = modifier

    def step(self, state, meta_state):
        """Apply rule to state, modifying meta_state."""
        del state
        self._modifier(meta_state)


class UpdateMetaStateValue(abstract_rule.AbstractRule):
    """Update a value in the meta_state dictionary.
    
    This rule assumes that the environment meta_state is a dictionary. It is
    commonly used as a one-time rule to indicate a change in task phase.
    """

    def __init__(self, key, value):
        """Constructor.

        Args:
            key: Key of the meta_state dictionary.
            value: Value to set as meta_state[key].
        """
        self._key = key
        self._value = value

    def step(self, state, meta_state):
        """Apply rule to state."""
        del state
        meta_state[self._key] = self._value



"""Gif writer to record a video while playing the demo.

Note: If the enter key prints `^M` instead of entering the input, run the
following command:
$ stty sane
"""

import imageio
import logging
import numpy as np
import os
import sys


class GifWriter(object):
    """GifWriter class.
    
    Usage:
        my_gif_writer = GifWriter('path/to/a/file.gif')
        for image in my_video:
            my_gif_writer.add(image)
        my_gif_writer.close()
    """

    def __init__(self, gif_file, fps=5):
        """Constructor.

        Args:
            gif_file: String. Full path to gif filename. Should end in '.gif'.
            fps: Int. Frames per second for the gif.
        """

        # If the gif directory does not exist, ask the user if they want to
        # create it.
        gif_file = os.path.expanduser(gif_file)
        gif_dir = os.path.dirname(gif_file)
        if not os.path.exists(gif_dir):
            print('Directory {} does not exist'.format(gif_dir))
            should_create = input(
                'Would you like to create that directory?  (y/n)')
            if should_create == 'y':
                print('Creating directory {}'.format(gif_dir))
                os.makedirs(gif_dir)
            else:
                print('exiting')
                sys.exit()

        # If the gif directory already exists, ask the user if they want to
        # override it.
        if os.path.isfile(gif_file):
            print('File {} to write gif to already exists.'.format(gif_file))
            should_override = input(
                'Would you like to override the file there?  (y/n)')
            if should_override == 'y':
                print('Removing {}'.format(gif_file))
                os.remove(gif_file)
            else:
                print('exiting')
                sys.exit()
                
        self._gif_file = gif_file
        self._images = []
        self._fps = fps

    def add(self, image):
        self._images.append(image)

    def close(self):
        """Write the gif."""
        print('Writing gif with {} images to file {}'.format(
            len(self._images), self._gif_file))
        imageio.mimsave(self._gif_file, self._images, fps=self._fps)



"""Rule that creates new sprites."""

from . import abstract_rule
import itertools


class CreateSprites(abstract_rule.AbstractRule):
    """Create new sprites and add them to a specified layer."""

    def __init__(self, layer, generator, without_overlapping=()):
        """Constructor.

        Args:
            layer: String. Must be a key in the environment state. Layer to
                which appearing sprites are added.
            generator: Callable returning a list of sprites. Must have
                'without_overlapping' kwarg. See
                ../state_initialization/sprite_generators.generate_sprites().
            without_overlapping: Iterable of strings. Each element must be a key
                in the environment state. Created sprites will not overlap
                sprites in these layers.
        """
        self._layer = layer
        self._generator = generator
        self._without_overlapping = without_overlapping

    def step(self, state, meta_state):
        """Apply rule to state."""
        del meta_state
        
        without_overlapping = list(itertools.chain(
            *[state[k] for k in self._without_overlapping]))
        new_sprite = self._generator(without_overlapping=without_overlapping)
        state[self._layer].extend(new_sprite)



"""Falling balls Task.

This task is not really a task, but instead used to demonstrate errors in
collision physics.

In this task some circles fall from the top of the screen, acted on by gravity.
The collide with each other, settling onto a wall at the bottom of the screen.
There is a divider in the bottom wall to add more collisions. The balls collide
with some elasticity with the walls and each other.

The collision dynamics for this task often do not look realistic. This is
because the collision force can be unrealistic under two conditions (both of
which are present in this task):
    * When an object collides while also accelerating due to some force at a
        distance.
    * When an object undergoes multiple collisions simultaneously.

The issues arise most often with sprites that have many vertices, such as the 
circles in this task.

Consider this task a word of caution when using collisions with many moving
objects and non-collision forces.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import observers
from moog import physics as physics_lib
from moog import sprite as sprite_lib
from moog import tasks
from moog.state_initialization import distributions as distribs
from moog.state_initialization import sprite_generators


def get_config(_):
    """Get environment config"""

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Ball generator
    ball_factors = distribs.Product(
        [distribs.Continuous('x', 0.25, 0.75),
         distribs.Continuous('y', 0.5, 0.9),
         distribs.Continuous('x_vel', -0.01, 0.01)],
        scale=0.1, shape='circle', c0=0, c1=0, c2=255, mass=1.,
    )
    ball_generator = sprite_generators.generate_sprites(
        ball_factors, num_sprites=4)

    # Walls
    bottom_wall = [[-1, 0.1], [2, 0.1], [2, -1], [-1, -1]]
    left_wall = [[0.05, -0.1], [0.05, 1.1], [-1, 1.1], [-1, -0.1]]
    right_wall = [[0.95, -0.1], [0.95, 1.1], [2, 1.1], [2, -0.1]]
    divider = [[0.45, -1], [0.45, 0.3], [0.55, 0.3], [0.55, -1]]
    walls = [
        sprite_lib.Sprite(shape=np.array(v), x=0, y=0, c0=128, c1=128, c2=128)
        for v in [bottom_wall, left_wall, right_wall, divider]
    ]

    def state_initializer():
        """Callable returning new state at each episode reset."""
        state = collections.OrderedDict([
            ('walls', walls),
            ('balls', ball_generator(disjoint=True)),
            ('agent', []),
        ])

        return state

    ############################################################################
    # Physics
    ############################################################################

    # Setting max_recursion_depth > 0 can increase stability
    # Setting update_angle_vel = False is recommended for stability
    collision = physics_lib.Collision(
        elasticity=0.6,
        symmetric=False,
        update_angle_vel=False,
        max_recursion_depth=2,
    )
    physics = physics_lib.Physics(
        (collision, 'balls', ['balls', 'walls']),
        (physics_lib.DownGravity(g=-0.001), 'balls'),
        updates_per_env_step=20,
    )
    
    ############################################################################
    # Task
    ############################################################################

    task = tasks.CompositeTask(timeout_steps=100)

    ############################################################################
    # Action space
    ############################################################################

    # Need an action space, so let it control an empty agent layer
    action_space = action_spaces.Grid(action_layers='agent')

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(image_size=(64, 64), anti_aliasing=1)

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': (),
    }
    return config



""".. include:: README.md"""

from .abstract_force import AbstractForce
from .abstract_force import AbstractNewtonianForce
from .abstract_physics import AbstractPhysics
from .collisions import Collision
from .constant_speed import ConstantSpeed
from .distance_fn_force import DistanceForce
from .distance_fn_force import linear_force_fn
from .distance_fn_force import spring_force_fn
from .friction import Drag
from .friction import KineticFriction
from .gravity import DownGravity
from .gravity import Gravity
from .maze_walk import DeterministicMazeWalk
from .maze_walk import RandomMazeWalk
from .maze_physics import MazePhysics
from .physics import Physics
from .random_force import RandomForce
from .tether_physics import Tether
from .tether_physics import TetherZippedLayers



"""Polygon generator functions."""

import numpy as np


def _polar2cartesian(r, theta):
  return r * np.array([np.cos(theta), np.sin(theta)])


def polygon(num_sides, theta_0=0.):
  """Generate the vertices of a regular polygon.

  Args:
    num_sides: Int. Number of sides of the polygon.
    theta_0: Float. Initial angle to start the vertices from.

  Returns:
    path: Array of vertices of the polygon, normalized so it has area 1.
  """
  theta = 2 * np.pi / num_sides
  path = np.array(
      [_polar2cartesian(1, i * theta + theta_0) for i in range(num_sides)])
  area = num_sides * np.sin(theta / 2) * np.cos(theta / 2)
  path = np.array(path) / np.sqrt(area)
  return path


def star(num_sides, point_height=1, theta_0=0.):
  """Generate the vertices of a regular star shape.

  Args:
    num_sides: Int. Number of sides (i.e. number of points) in the star.
    point_height: Scalar. Height of each point of the star, relative to the
      radius of the star's inscribed circle.
    theta_0: Float. Initial angle to start the vertices from.

  Returns:
    path: Array of vertices of the star, normalized so the star has area 1.
  """
  point_to_center = 1 + point_height
  theta = 2 * np.pi / num_sides
  path = np.empty([2 * num_sides, 2])
  for i in range(num_sides):
    path[2 * i] = _polar2cartesian(1, i * theta + theta_0)
    path[2 * i + 1] = _polar2cartesian(point_to_center,
                                       (i + 0.5) * theta + theta_0)

  area = point_to_center * num_sides * np.sin(theta / 2)
  path = np.array(path) / np.sqrt(area)
  return path


def spokes(num_sides, spoke_height=1, theta_0=0.):
  """Generate the vertices of a regular rectangular spoke shape.

  This is like a star, except the points are rectangular. For example, if
  num_sides = 4, it will look like this:

                            O       O
                          O   O   O   O
                        O       O       O
                          O           O
                            O       O
                          O           O
                        O       O       O
                          O   O   O   O
                            O       O

  Args:
    num_sides: Int. Number of sides (i.e. number of points) in the star.
    spoke_height: Scalar. Height of each spoke, relative to the radius of the
      spoke shape's inscribed circle.
    theta_0: Float. Initial angle to start the vertices from.

  Returns:
    path: Array of vertices of the spoke shape, normalized so the spoke shape
      has area 1.
  """
  theta = 2 * np.pi / num_sides
  path = np.empty([3 * num_sides, 2])
  spoke = _polar2cartesian(spoke_height, -0.5 * theta + theta_0)
  for i in range(num_sides):
    vertex = _polar2cartesian(1, i * theta + theta_0)
    path[3 * i] = spoke + vertex
    path[3 * i + 1] = vertex
    spoke = _polar2cartesian(spoke_height, (i + 0.5) * theta + theta_0)
    path[3 * i + 2] = spoke + vertex

  area = num_sides * np.sin(theta / 2) * (2 + np.cos(theta / 2))

  path = np.array(path) / np.sqrt(area)
  return path


"""Constant speed corrective physics.

The ConstantSpeed class is a physics object that forces sprites in specified
layers to have constant speed. It is typically used as a corrective physics.
"""

from . import abstract_physics
import itertools
import numpy as np


class ConstantSpeed(abstract_physics.AbstractPhysics):
    """Constant speed physics class.
    
    Forces sprites in specified layers to have constant speed by normalizing
    their velocities.
    """

    def __init__(self, layer_names, speed):
        """Constructor.

        Args:
            layer_names: String or iterable of strings corresponding to layers
                in environment state. All sprites in these layers will have
                speed equal to `speed` argument.
            speed: Float. Sprite velocities will be normalized to have norm
                equal to this (or zero if their velocity is zero).
        """
        if not isinstance(layer_names, (list, tuple)):
            layer_names = [layer_names]
        self._layer_names = layer_names
        self._speed = speed

    def apply_physics(self, state, updates_per_env_step):
        """Apply the physics.
        
        Args:
            state: Environment state.
            updates_per_env_step: Int. Number of times the physics is applied
                per environment step.
        """
        sprites = [s for k in self._layer_names for s in state[k]]
        for s in sprites:
            norm_velocity = np.linalg.norm(s.velocity)
            if norm_velocity:
                s.velocity = self._speed * s.velocity / norm_velocity



"""Abstract observer."""

import abc


class AbstractObserver(abc.ABC):
    """Abstract observer class.
    
    All observers must inherit from this class.
    """
    
    @abc.abstractmethod
    def __call__(self, state):
        """Observe the environment state.
        
        Args:
            state: OrderedDict of iterables of sprites.

        Returns:
            observation. Observation of the environment state. Type and size of
                the observation may vary over time, depending on the observer.
        """
        pass

    @abc.abstractmethod
    def observation_spec(self):
        """Get observation spec for the output.
        
        Returns:
            dm_env.specs.ArraySpec or nested structure of such.
        """
        pass



"""Restore and plot logged data.

This file can be used to visualize data logged by
..moog.env_wrappers.logger.LoggingEnvironment.

To run, first play .run_demo.py with flag --log_data=True. Then find the log
directory that was written to, which will be printed at the beginning of playing
the demo and should be of the form 'logs/config_name/$integer/$date_time'. Then
using that directory, run
$ python3 restore_logged_data.py --config='path.to.your.config' \
  --log_directory=your_log_directory

You may need to also set the --level flag.
"""
import sys
sys.path.insert(0, '..')

from absl import app
from absl import flags

import collections
import importlib
import json
import logging
import numpy as np
import os
from moog import observers
from moog import sprite as sprite_lib

from matplotlib import path as mpl_path
from matplotlib import pyplot as plt
from matplotlib import transforms as mpl_transforms

FLAGS = flags.FLAGS
flags.DEFINE_string('config',
                    'example_configs.pong',
                    'Filename of task config to use.')
flags.DEFINE_integer('level', 0, 'Level of task config to run.')
flags.DEFINE_string('log_directory',
                    'logs/pong/0/2021_02_06_16_18_58',
                    'Directory of logs to restore.')
flags.DEFINE_integer('num_episodes', 3, 'Number of episodes to restore.')


def _create_new_sprite(sprite_kwargs, vertices=None):
    """Create new sprite from factors.

    Args:
        sprite_kwargs: Dict. Keyword arguments for sprite_lib.Sprite.__init__().
            All of the strings in sprite_lib.Sprite.FACTOR_NAMES must be keys of
            sprite_kwargs.
        vertices: Optional numpy array of vertices. If provided, are used to
            define the shape of the sprite. Otherwise, sprite_kwargs['shape'] is
            used.

    Returns:
        Instance of sprite_lib.Sprite.
    """
    if vertices is not None:
        # Have vertices, so must invert the translation, rotation, and
        # scaling transformations to get the original sprite shape.
        center_translate = mpl_transforms.Affine2D().translate(
            -sprite_kwargs['x'], -sprite_kwargs['y'])
        x_y_scale = 1. / np.array([
            sprite_kwargs['scale'],
            sprite_kwargs['scale'] * sprite_kwargs['aspect_ratio']
        ])
        transform = (
            center_translate +
            mpl_transforms.Affine2D().rotate(-sprite_kwargs['angle']) +
            mpl_transforms.Affine2D().scale(*x_y_scale)
        )
        vertices = mpl_path.Path(vertices)
        vertices = transform.transform_path(vertices).vertices

        sprite_kwargs['shape'] = vertices

    return sprite_lib.Sprite(**sprite_kwargs)


def _state_str_to_image(state_str, renderer, attributes, stored_sprites):
    """Convert state string to image.

    Args:
        state_str: String. This should be a state string from an episode log.
            It should contain a list of layers, each of which is of the form
            [layer_name, sprites_attributes] where sprites_attributes is a list
            of attributes of all the sprites in the layer.
        renderer: Instance of observers.PILRenderer. Used to renderer the state.
        attributes: List of strings. Each must be either an element of
            sprite_lib.Sprite.FACTOR_NAMES of 'id'.
        stored_sprites: Dict. Keys are unique id values of sprites, and values
            are sprite instances. This dictionary is dynamically updated to keep
            track of sprites so they don't need to be re-instantianted every
            step.

    Returns:
        Image of rendered state.
    """
    state = collections.OrderedDict()

    # Will keep track of which sprite ids are still in use
    active_sprite_ids = []

    for layer_name, layer_str in state_str:
        layer_sprites = []

        for sprite_str in layer_str:
            create_new_sprite = False
            vertices = None

            if len(sprite_str) == len(attributes) + 1:
                # Vertices are the last element of sprite_str
                vertices = np.array(sprite_str.pop(-1))
                create_new_sprite = True
            elif len(sprite_str) != len(attributes):
                raise ValueError(
                    'len(sprite_str) = {} must be equal to or one greater than '
                    'len(attributes) = {}'.format(
                        len(sprite_str), len(attributes)))

            # Kwargs for the constructor of sprite_lib.Sprite()
            sprite_kwargs = {
                k: v for k, v in zip(attributes, sprite_str)
            }

            # All attributes other than 'id' should be in
            # sprite_lib.Sprite.FACTOR_NAMES
            sprite_id = sprite_kwargs.pop('id')
            active_sprite_ids.append(sprite_id)

            if sprite_id not in stored_sprites:
                create_new_sprite = True

            # Create new sprite if necessary, else update old sprite
            if create_new_sprite:
                sprite = _create_new_sprite(sprite_kwargs, vertices=vertices)
                stored_sprites[sprite_id] = sprite
            else:
                sprite = stored_sprites[sprite_id]
                sprite_lib.update_sprite(sprite, **sprite_kwargs)

            layer_sprites.append(sprite)
        state[layer_name] = layer_sprites

    # Purge stored_sprites, removing all inactive sprites
    inactive_ids = [k for k in stored_sprites if k not in active_sprite_ids]
    [stored_sprites.pop(k) for k in inactive_ids]

    return renderer(state)


def main(_):
    """Restore and render logged data."""
    log_dir = FLAGS.log_directory

    ############################################################################
    #### Create renderer
    ############################################################################

    # Use the PILRenderer in config['observers'] if it exists
    config = importlib.import_module(FLAGS.config)
    config = config.get_config(FLAGS.level)
    renderer_observer = False
    for renderer in config['observers'].values():
        if isinstance(renderer, observers.PILRenderer):
            renderer_observer = True
            break
    if not renderer_observer:
        # config['observers'] has no PILRenderer, so use this one and hope for
        # the best
        renderer = observers.PILRenderer(
            image_size=(256, 256),
            anti_aliasing=1,
            color_to_rgb='hsv_to_rgb',  # Depends on the config
        )

    logging.info('Renderer instantiated.')

    ############################################################################
    #### Load logged episodes
    ############################################################################

    log_filenames = sorted(
        filter(lambda s: s.isnumeric(), os.listdir(log_dir)))
    log_filenames = log_filenames[:min(FLAGS.num_episodes, len(log_filenames))]
    episode_strings = [
        json.load(open(os.path.join(log_dir, x))) for x in log_filenames]
    attributes = json.load(open(os.path.join(log_dir, 'attributes.txt')))

    logging.info('Episode strings read.')

    ############################################################################
    #### Render logged episodes
    ############################################################################

    episode_images = []
    stored_sprites = {}
    for i, episode_str in enumerate(episode_strings):
        logging.info('Rendering episode {}'.format(i))
        for timestep in episode_str:
            state_str = timestep[-1]
            image = _state_str_to_image(
                state_str, renderer, attributes, stored_sprites)
            episode_images.append(image)

    ############################################################################
    #### Display video of logged episodes
    ############################################################################

    logging.info('Displaying rendered episodes in infinite loop.')
    ax = plt.subplots()[1]
    imshow = ax.imshow(episode_images[0])

    index = 0
    while True:
        index = (index + 1) % len(episode_images)
        imshow.set_data(episode_images[index])
        plt.draw()
        plt.pause(0.001)


if __name__ == "__main__":
    app.run(main)



""".. include:: README.md"""

from .abstract_wrapper import AbstractEnvironmentWrapper
from .logger import LoggingEnvironment
from .multi_agent import MultiAgentEnvironment
from .simulation import SimulationEnvironment



""".. include:: README.md"""

from .abstract_task import AbstractTask
from .composite_task import CompositeTask
from .contact_reward import ContactReward
from .reset import Reset
from .stay_alive import StayAlive



"""Abstract force classes.

This file contains AbstractForce and AbstraceNewtonianForce classes.
"""

import abc
import numpy as np


class AbstractForce(abc.ABC):
    """Abstract force class.
    
    All physics forces must inherit from this class.
    """
    
    @abc.abstractmethod
    def step(self, *sprites, updates_per_env_step):
        """Step the force.
        
        Args:
            *sprites: Iterable of sprites on which the force operates.
            updates_per_env_step: Int. Number of times this force step is called
                for each step of the physics in the environment. This is used
                for example to scale down Netwonian forces so that from the
                user's perspective the force is invariant to
                updates_per_env_step.
        """
        pass

    def reset(self, state):
        """Reset the force.

        This is called at every episode reset of the environment. Usually it
        does nothing (usually the forces don't change between episodes), but
        for maze environments this is used to initialize the maze for the
        episode.

        Args:
            state: Environment state.
        """
        pass


class AbstractNewtonianForce(AbstractForce, metaclass=abc.ABCMeta):
    """Abstract Newtonian force class.
    
    Forces abiding by F = ma should inherit from this class.
    """

    @abc.abstractmethod
    def _compute_forces(self, *sprites):
        """Compute forces on all sprites.
        
        Args:
            *sprites: Iterable of sprites on which the force operates.
        
        Returns:
            Iterable with same length as sprites. Each element is a scalar numpy
            array [f_x, f_y], the force to be applied to the corresponding
            sprite.
        """
        pass
    
    def step(self, *sprites, updates_per_env_step):
        forces = self._compute_forces(*sprites)
        for sprite, force in zip(sprites, forces):
            if not np.isfinite(sprite.mass):
                # Good to catch this because sometimes we might make a sprite
                # have infinite mass to prevent it from moving, but we don't
                # want that sprite's velocity and consequently position to
                # become NaN.
                continue
            delta_vel = force / (sprite.mass * float(updates_per_env_step))
            sprite.velocity += delta_vel



# Copyright 2019 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Wrapper to make object-oriented games conform to the OpenAI Gym interface.

Note: This wrapper does not inherit from
abstract_wrapper.AbstractEnvironmentWrapper, because unlike other wrappers this
one (intentionally) changes the API of the environment.
"""

from dm_env import specs
from gym import spaces
import numpy as np


def _spec_to_space(spec):
    """Convert dm_env.specs to gym.Spaces."""
    if isinstance(spec, list):
        return spaces.Tuple([_spec_to_space(s) for s in spec])
    elif isinstance(spec, specs.DiscreteArray):
        return spaces.Discrete(spec.num_values)
    elif isinstance(spec, specs.BoundedArray):
        return spaces.Box(
            spec.minimum.item(),
            spec.maximum.item(),
            shape=spec.shape,
            dtype=spec.dtype)
    elif isinstance(spec, dict):
        print("not here")
        pass
    else:
        raise ValueError('Unknown type for specs: {}'.format(spec))


class GymWrapper(object):
    """Wraps a object-oriented game environment into a Gym interface.

    Observations will be a dictionary, with the same keys as the 'observers'
    dict provided when constructing a object-oriented game environment.
    Rendering is always performed, so calling render() is a no-op.
    """
    metadata = {'render.modes': ['rgb_array']}

    def __init__(self, env):
        self._env = env
        self._last_render = None
        self._action_space = None
        self._observation_space = None

        # Reset object-oriented to setup the observation_specs correctly
        self._env.reset()

    @property
    def observation_space(self):
        if self._observation_space is None:
            components = {}
            for key, value in self._env.observation_spec().items():
                if value.dtype == np.uint8:
                    components[key] = spaces.Box(
                        0, 255, value.shape, dtype=value.dtype)
                else:
                    components[key] = spaces.Box(
                        -np.inf, np.inf, value.shape, dtype=value.dtype)
            self._observation_space = spaces.Dict(components)
        return self._observation_space

    @property
    def action_space(self):
        if self._action_space is None:
            self._action_space = _spec_to_space(self._env.action_spec())
        return self._action_space

    def _process_obs(self, obs):
        """Convert and processes observations."""
        for k, v in obs.items():
            obs[k] = np.asarray(v)
            if obs[k].dtype == bool:
                # Convert boolean 'success' into an float32 to predict it.
                obs[k] = obs[k].astype(np.float32)
            if k == 'image':
                self._last_render = obs[k]

        return obs

    def step(self, action):
        """Main step function for the environment.

        Args:
            action: Array R^4

        Returns:
            obs: dict of observations. Follows from the 'renderers'
                configuration
                provided as parameters to object-oriented games.
            reward: scalar reward.
            done: True if terminal state.
            info: dict with extra information (e.g. discount factor).
        """
        time_step = self._env.step(action)
        obs = self._process_obs(time_step.observation)
        reward = time_step.reward or 0
        done = time_step.last()
        info = {'discount': time_step.discount}
        return obs, reward, done, info

    def reset(self):
        """Reset environment.

        Returns:
            obs: dict of observations. Follows from the 'renderers'
                configuration provided as parameters to object-oriented games.
        """
        time_step = self._env.reset()
        return self._process_obs(time_step.observation)

    def render(self, mode='rgb_array'):
        """Render function, noop for compatibility.

        Args:
            mode: unused, always returns an RGB array.

        Returns:
            Last RGB observation (cached from last observation with key
                'image').
        """
        del mode
        return self._last_render

    def close(self):
        """Unused."""
        pass



"""Installation script."""

from setuptools import find_packages
from setuptools import setup
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='moog-games',
    version='1.6',
    description=(
        'Modular object oriented games utils is a python-based game engine.'),
    author='Nicholas Watters',
    url='https://github.com/jazlab/moog.github.io',
    license='MIT license',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=[
        'ai',
        'reinforcement-learning',
        'python',
        'machine learning',
        'objects',
        'psychology',
        'neurophysiology',
        'psychophysics',
        'physics',
        'environment',
    ],
    packages=(
        ['moog_demos', 'moog'] + 
        ['moog_demos.' + x for x in find_packages('moog_demos')] +
        ['moog.' + x for x in find_packages('moog')]
    ),
    install_requires=[
        'cmake'
        'numpy==2.2.3',
        'matplotlib==3.10.0',
        'imageio==2.37.0',
        'mss==10.0.0',
        'tqdm==4.67.1',
        'dm_env==1.6',
        'gym==0.26.2',
    ],
    tests_require=[
        'nose==1.3.7',
        'pytest==8.3.4',
    ],
    test_suite='nose.collector',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering',
    ],
)


# This file was forked and modified from the file here:
# https://github.com/deepmind/spriteworld/blob/master/spriteworld/sprite_generators.py
# Here is the license header for that file:

# Copyright 2019 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Generators for producing lists of sprites based on factor distributions."""

import itertools
import numpy as np
from moog import sprite


def generate_sprites(factor_dist,
                     num_sprites=1,
                     max_recursion_depth=int(1e4),
                     fail_gracefully=False):
    """Create callable that samples sprites from a factor distribution.

    Example usage:
        ```python
        sprite_factors = distribs.Product(
            [distribs.Continuous('x', 0.2, 0.8),
             distribs.Continuous('y', 0.2, 0.8),
             distribs.Continuous('x_vel', -0.03, 0.03),
             distribs.Continuous('y_vel', -0.03, 0.03)],
            shape='circle, scale=0.1, c0=255, c1=0, c2=0,
        )
        sprite_gen = sprite_generators.generate_sprites(
            sprite_factors, num_sprites=lambda: np.random.randint(3, 6))
        
        def _state_initializer():
            ...
            other_sprites = ...
            ...
            sprites = sprite_gen(
                disjount=True, without_overlapping=other_sprites)
            state = collections.OrderedDict([
                ('other_sprites', other_sprites),
                ('sprites', sprites),
            ])
        ```

    Args:
        factor_dist: The factor distribution from which to sample. Should be an
            instance of spriteworld.factor_distributions.AbstractDistribution.
        num_sprites: Int or callable returning int. Number of sprites to
            generate per call.
        max_recursion_depth: Int. Maximum recursion depth when rejection
            sampling to generate sprites without overlap.
        fail_gracefully: Bool. Whether to return a list of sprites or raise
            RecursionError if max_recursion_depth is exceeded.

    Returns:
        _generate: Callable that returns a list of Sprites.
    """
    def _overlaps(s, other_sprites):
        """Whether s overlaps any sprite in other_sprites."""
        if len(other_sprites) == 0:
            return False
        else:
            overlaps = [s.overlaps_sprite(x) for x in other_sprites]
            return any(overlaps)

    def _generate(disjoint=False, without_overlapping=[]):
        """Return a list of sprites.
        
        Args:
            disjoint: Boolean. If true, all generated sprites will be disjoint.
            without_overlapping: Optional iterable of ../sprite/Sprite
                instances. If specified, all generated sprites will not overlap
                any sprites in without_overlapping.
        """
        n = num_sprites() if callable(num_sprites) else num_sprites
        sprites = []
        for _ in range(n):
            s = sprite.Sprite(**factor_dist.sample())
            count = 0
            while _overlaps(s, without_overlapping):
                if count > max_recursion_depth:
                    if fail_gracefully:
                        return sprites
                    else:
                        raise RecursionError(
                            'max_recursion_depth exceeded trying to initialize '
                            'a non-overlapping sprite.')
                count += 1
                s = sprite.Sprite(**factor_dist.sample())
            sprites.append(s)
            if disjoint:
                without_overlapping = without_overlapping + [s]
        
        return sprites

    return _generate


def chain_generators(*sprite_generators):
    """Chain generators by concatenating output sprite sequences.

    Essentially an 'AND' operation over sprite generators. This is useful when
    one wants to control the number of samples from the modes of a multimodal
    sprite distribution.

    Note that factor_distributions.Mixture provides weighted mixture
    distributions, so chain_generators() is typically only used when one wants
    to forces the different modes to each have a non-zero number of sprites.

    Args:
        *sprite_generators: Callable sprite generators.

    Returns:
        _generate: Callable returning a list of sprites.
    """

    def _generate(*args, **kwargs):
        return list(itertools.chain(*[generator(*args, **kwargs)
                                      for generator in sprite_generators]))

    return _generate


def sample_generator(sprite_generators, p=None):
    """Sample one element from a set of sprite generators.

    Essential an 'OR' operation over sprite generators. This returns a callable
    that samples a generator from sprite_generators and calls it.

    Note that if sprite_generators each return 1 sprite, this functionality can
    be achieved with spriteworld.factor_distributions.Mixture, so
    sample_generator is typically used when sprite_generators each return
    multiple sprites. Effectively it allows dependant sampling from a multimodal
    factor distribution.

    Args:
        sprite_generators: Iterable of callable sprite generators.
        p: Probabilities associated with each generator. If None, assumes
            uniform distribution.

    Returns:
        _generate: Callable sprite generator.
    """

    def _generate(*args, **kwargs):
        sampled_generator = np.random.choice(sprite_generators, p=p)
        return sampled_generator(*args, **kwargs)

    return _generate


def shuffle(sprite_generator):
    """Randomize the order of sprites sample from sprite_generator.

    This is useful because sprites are z-layered with occlusion according to
    their order, so if sprite_generator is the output of chain_generators(),
    then sprites from some component distributions will always be behind sprites
    from others.

    An alternate design would be to let the environment handle sprite ordering,
    but this design is preferable because the order can be controlled more
    finely. For example, this allows the user to specify one sprite (e.g. the
    agent's body) to always be in the foreground while all the others are
    randomly ordered.

    Args:
        sprite_generator: Callable return a list of sprites.

    Returns:
        _generate: Callable sprite generator.
    """

    def _generate(*args, **kwargs):
        sprites = sprite_generator(*args, **kwargs)
        order = np.arange(len(sprites))
        np.random.shuffle(order)
        return [sprites[i] for i in order]

    return _generate



"""Red-Green Task.

This task is a variant of the task used in this paper:
Smith, K. A., Peres, F., Vul, E., & Tenebaum, J. (2017). Thinking inside the
box: Motion prediction in contained spaces uses simulation. In CogSci.

In this task, there is a blue ball that bounces in an enclosed rectangular
arena. The arena may have gray rectangular obstacles that the blue ball bounces
off. The arena has one green box and one red box. The subject's goal is to
predict which of the green/red boxes the blue ball will contact first.

In this particular implementation, the subject moves a token at the bottom of
the screen left or right to indicate its choice.

The main entrypoint is the get_config() function at the bottom of this file.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import game_rules
from moog import observers
from moog import physics as physics_lib
from moog import sprite
from moog import tasks
from moog.state_initialization import distributions as distribs
from moog.state_initialization import sprite_generators


class RadialVelocity(distribs.AbstractDistribution):
    """Radial velocity distribution."""

    def __init__(self, speed):
        """Constructor.

        Args:
            speed: Float. Speed of the sampled velocities.
        """
        self._speed = speed

    def sample(self, rng):
        rng = self._get_rng(rng)
        theta = rng.uniform(0., 2 * np.pi)
        x_vel = self._speed * np.cos(theta)
        y_vel = self._speed * np.sin(theta)
        return {'x_vel': x_vel, 'y_vel': y_vel}

    def contains(self, spec):
        if 'x_vel' not in spec or 'y_vel' not in spec:
            return False

        vel_norm = np.linalg.norm([spec['x_vel'], spec['y_vel']])
        if np.abs(vel_norm - self._speed) < _EPSILON:
            return True
        else:
            return False

    def to_str(self, indent):
        s = 'RadialVelocity({})'.format(self._speed)
        return indent * '  ' + s

    @property
    def keys(self):
        return set(['x_vel', 'y_vel'])
        

def _get_config(num_obstacles, valid_step_range):
    """Get environment config.
    
    Args:
        num_obstacles: Int. Number of obstacles.
        valid_step_range: 2-iterable of ints. (min_num_steps, max_num_steps).
            All trials must have duration in this step range.
    
    Returns:
        config: Config dictionary to pass to environment constructor.
    """

    ############################################################################
    # Physics
    ############################################################################

    elastic_collision = physics_lib.Collision(
        elasticity=1., symmetric=False, update_angle_vel=False)
    physics = physics_lib.Physics(
        (elastic_collision, 'ball', 'walls'),
        updates_per_env_step=10,
    )

    def _predict_trial_end(state):
        """Predict whether a trial will end in step range and true response.

        Args:
            state: OrderedDict of sprite layers. Initial state of environment.
        
        Returns:
            valid_trial: Bool. Whether trial will end with number of steps in
                valid_step_range.
            contact_color: Binary. 0 if ball will contact red first, 1 if it
                will contact green first.
        """
        for step in range(valid_step_range[1]):
            red_overlap = state['ball'][0].overlaps_sprite(state['red'][0])
            green_overlap = state['ball'][0].overlaps_sprite(state['green'][0])
            if red_overlap or green_overlap:
                if step < valid_step_range[0]:
                    return False, None
                else:
                    contact_color = 0 if red_overlap else 1
                    return True, contact_color
            physics.step(state)
        return False, None
    
    ############################################################################
    # Sprite initialization
    ############################################################################

    # Ball generator
    ball_factors = distribs.Product(
        [distribs.Continuous('x', 0.15, 0.85),
         distribs.Continuous('y', 0.15, 0.85),
         RadialVelocity(speed=0.03)],
        scale=0.05, shape='circle', c0=64, c1=64, c2=255,
    )
    ball_generator = sprite_generators.generate_sprites(
        ball_factors, num_sprites=1, max_recursion_depth=100,
        fail_gracefully=True)

    # Obstacle generator
    obstacle_factors = distribs.Product(
        [distribs.Continuous('x', 0.2, 0.8),
         distribs.Continuous('y', 0.2, 0.8)],
        scale=0.2, shape='square', c0=128, c1=128, c2=128,
    )
    obstacle_generator = sprite_generators.generate_sprites(
        obstacle_factors, num_sprites=2 + num_obstacles,
        max_recursion_depth=100, fail_gracefully=True)

    # Walls
    bottom_wall = [[-1, 0.1], [2, 0.1], [2, -1], [-1, -1]]
    top_wall = [[-1, 0.95], [2, 0.95], [2, 2], [-1, 2]]
    left_wall = [[0.05, -1], [0.05, 4], [-1, 4], [-1, -1]]
    right_wall = [[0.95, -1], [0.95, 4], [2, 4], [2, -1]]
    walls = [
        sprite.Sprite(shape=np.array(v), x=0, y=0, c0=128, c1=128, c2=128)
        for v in [bottom_wall, top_wall, left_wall, right_wall]
    ]

    def state_initializer():
        """Callable returning new state at each episode reset."""
        obstacles = obstacle_generator(disjoint=True)
        ball = ball_generator(without_overlapping=obstacles)
        if len(obstacles) < num_obstacles + 2 or not ball:
            # Max recursion depth failed trying to generate without overlapping
            return state_initializer()
        
        red = obstacles[0]
        green = obstacles[1]
        obstacles = obstacles[2:]

        # Set the colors of the red and green boxes
        red.c0 = 255
        red.c1 = 0
        red.c2 = 0
        green.c0 = 0
        green.c1 = 255
        green.c2 = 0

        # Create agent and response tokens at the bottom of the sreen
        agent = sprite.Sprite(x=0.5, y=0.06, shape='spoke_4', scale=0.03, c0=255, c1=255, c2=255)
        responses = [
            sprite.Sprite(x=0.6, y=0.06, shape='square', scale=0.03, c0=255, c1=0, c2=0),
            sprite.Sprite(x=0.4, y=0.06, shape='square', scale=0.03, c0=0, c1=255, c2=0),
        ]

        state = collections.OrderedDict([
            ('walls', walls + obstacles),
            ('red', [red]),
            ('green', [green]),
            ('ball', ball),
            ('responses', responses),
            ('agent', [agent]),
        ])

        # Rejection sampling if trial won't finish in valid step range
        original_ball_position = np.copy(ball[0].position)
        original_ball_velocity = np.copy(ball[0].velocity)
        valid_trial, contact_red = _predict_trial_end(state)
        if valid_trial:
            ball[0].position = original_ball_position
            ball[0].velocity = original_ball_velocity
            agent.metadata = {'true_contact_color': contact_red}
        else:
            return state_initializer()

        return state

    ############################################################################
    # Task
    ############################################################################

    def _reward_fn(sprite_agent, sprite_response):
        response_green = sprite_response.c0 < 128
        if sprite_agent.metadata['true_contact_color'] == response_green:
            return 1.
        else:
            return -1.
            
    contact_reward = tasks.ContactReward(
        reward_fn=_reward_fn,
        layers_0='agent',
        layers_1='responses',
        reset_steps_after_contact=10,
    )
    task = tasks.CompositeTask(contact_reward, timeout_steps=400)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Grid(
        scaling_factor=0.015,
        action_layers='agent',
        control_velocity=True,
    )

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(image_size=(64, 64), anti_aliasing=1)

    ############################################################################
    # Game rules
    ############################################################################

    # Stop ball on contact with red or green box
    def _stop_ball(s):
        s.velocity = np.zeros(2)
    stop_ball = game_rules.ModifyOnContact(
        layers_0='ball',
        layers_1=('red', 'green'),
        modifier_0=_stop_ball
    )

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': (stop_ball,),
    }
    return config


def get_config(level):
    """Get config to pass to environment constructor.

    Args:
        level: Int. Number of obstacles in arena.
    """
    if not isinstance(level, int):
        raise ValueError(f'level is {level}, but must be an integer.')
    return _get_config(num_obstacles=level, valid_step_range=(50, 150))



"""Match-to-sample task.

In this task there are some colored circles (targets) on a ring, and an agent
avatar in the center of the ring. After an initial stimulus period where the
colored targets are visible, they all turn grey and begin rotating together.
After some time they stop and a colored cue appears on the agent avatar. The
subject must identify the target that had the same color as the cue and respond
by navigating towards that target.

This task requires workimg memory of multiple objects with features (colors).
"""

import collections
import copy
import functools
import numpy as np

from moog import action_spaces
from moog import game_rules as gr
from moog import observers
from moog import physics as physics_lib
from moog import shapes
from moog import sprite
from moog import tasks
from moog.state_initialization import distributions as distribs


def _get_polygon(num_sides, min_angle):
    """Get polygon vertices centered around 0 with radius 1."""
    # Get vertex angles
    angles = [0.]
    def _get_random_angle():
        return np.random.uniform(min_angle, 2 * np.pi - min_angle)
    for _ in range(num_sides - 1):
        valid_angle = False
        while not valid_angle:
            angle = _get_random_angle()
            if all([np.abs(angle - a) > min_angle for a in angles]):
                valid_angle = True
        angles.append(angle)

    # Sort and randomly rotate the angles
    angles = np.sort(angles)
    angles += np.random.uniform(0., 2 * np.pi)

    # Generate the vertices
    vertices = np.stack((np.sin(angles), np.cos(angles)), axis=1)

    return vertices


class BeginMotion(gr.AbstractRule):
    """Custom rule for motion initiation."""
    
    def __init__(self, angle_vel_range):
        """Constructor.

        Args:
            angle_vel_range: Tuple of non-negative floats. Range of angular
                velocity magnitudes, in radians per step.
        """
        self._angle_vel_range = angle_vel_range

    def step(self, state, meta_state):
        del meta_state
        targets = state['targets']
        covers = state['covers']
        angle_vel = np.random.uniform(*self._angle_vel_range)
        angle_vel *= (2 * np.random.randint(2) - 1)
        for t, c in zip(targets, covers):
            relative_pos = t.position - 0.5
            perpendicular = np.matmul(np.array([[0, -1], [1, 0]]), relative_pos)
            radius = np.linalg.norm(relative_pos)
            vel = perpendicular * radius * angle_vel
            t.velocity = vel
            c.velocity = vel


def get_config(num_targets):
    """Get environment config.
    
    Args:
        num_targets: Int. Number of targets.
    """
    if num_targets == 0 or not isinstance(num_targets, int):
        raise ValueError(
            f'num_targets is {num_targets}, but must be a positive integer')

    ############################################################################
    # State initialization
    ############################################################################

    screen = sprite.Sprite(
        x=0.5, y=0.5, shape='square', scale=2., c0=0.6, c1=0.7, c2=0.7)
        
    target_factor_distrib = distribs.Product(
        [distribs.Continuous('c0', 0., 1.)],
        shape='circle', scale=0.085, c1=1., c2=1.,
    )
    cover_factors = dict(
        mass=0., shape='circle', scale=0.1, c0=0., c1=0., c2=0.5, opacity=0)
    
    def state_initializer():
        """State initializer method to be fed into environment."""

        # Get targets and covers
        sprite_positions = 0.5 + 0.35 * _get_polygon(num_targets, 0.7)
        target_factors = [
            target_factor_distrib.sample() for _ in range(num_targets)]
        targets = [
            sprite.Sprite(x=pos[0], y=pos[1], **factors)
            for pos, factors in zip(sprite_positions, target_factors)
        ]
        covers = [
            sprite.Sprite(x=pos[0], y=pos[1], **cover_factors)
            for pos in sprite_positions
        ]

        # Tag the cover metadata based on whether they are prey or not
        for i, s in enumerate(covers):
            if i == 0:
                s.metadata = {'prey': True}
            else:
                s.metadata = {'prey': False}
        
        # Make cue have the same factors as the first target, except slightly
        # smaller
        cue_factors = copy.deepcopy(target_factors[0])
        cue_factors['scale'] = 0.7 * target_factors[0]['scale']
        cue = sprite.Sprite(
            x=0.5, y=0.501, opacity=0, mass=np.inf, **cue_factors)

        agent = sprite.Sprite(
            x=0.5, y=0.5, shape='circle', scale=0.1, c0=0.4, c1=0., c2=1.,
            mass=np.inf)
        annulus_verts = shapes.annulus_vertices(0.34, 0.36)
        annulus = sprite.Sprite(
            x=0.5, y=0.5, shape=annulus_verts, scale=1., c0=0., c1=0., c2=0.3)
        
        state = collections.OrderedDict([
            ('annulus', [annulus]),
            ('targets', targets),
            ('covers', covers),
            ('agent', [agent]),
            ('cue', [cue]),
            ('screen', [screen]),
        ])
        return state

    ################################################################################
    # Physics
    ################################################################################

    drag = (physics_lib.Drag(coeff_friction=0.25), ['agent', 'cue'])
    tether_covers = physics_lib.TetherZippedLayers(
        ('targets', 'covers'), anchor=np.array([0.5, 0.5]))
    physics = physics_lib.Physics(
        drag,
        updates_per_env_step=1,
        corrective_physics=[tether_covers],
    )

    ################################################################################
    # Task
    ################################################################################

    contact_task = tasks.ContactReward(
        reward_fn=lambda _, s: 1 if s.metadata['prey'] else -1,
        layers_0='agent',
        layers_1='covers',
    )

    def _should_reset(state, meta_state):
        should_reset = (
            state['covers'][0].opacity == 0 and
            meta_state['phase'] == 'response'
        )
        return should_reset
    reset_task = tasks.Reset(
        condition=_should_reset,
        steps_after_condition=15,
    )

    task = tasks.CompositeTask(contact_task, reset_task, timeout_steps=800)
    
    ################################################################################
    # Action Space
    ################################################################################

    action_space = action_spaces.Joystick(
        scaling_factor=0.01, action_layers=['agent', 'cue'])

    ################################################################################
    # Observer
    ################################################################################

    _polygon_modifier = observers.polygon_modifiers.FirstPersonAgent(
        agent_layer='agent')
    observer = observers.PILRenderer(
        image_size=(64, 64),
        anti_aliasing=1,
        color_to_rgb='hsv_to_rgb',
        polygon_modifier=_polygon_modifier,
    )

    ############################################################################
    # Game rules
    ############################################################################

    def _make_opaque(s):
        s.opacity = 255
    def _make_transparent(s):
        s.opacity = 0

    # Screen Phase

    screen_phase = gr.Phase(duration=1, name='screen')

    # Visible Phase

    disappear_screen = gr.ModifySprites('screen', _make_transparent)
    visible_phase = gr.Phase(
        one_time_rules=disappear_screen, duration=2, name='visible')

    # Motion Phase

    def _move(s):
        s.velocity = np.random.uniform(-0.25, 0.25, size=(2,))

    cover_targets = gr.ModifySprites('covers', _make_opaque)
    begin_motion = BeginMotion(angle_vel_range=(0.1, 0.3))
    motion_phase = gr.Phase(
        one_time_rules=[cover_targets, begin_motion],
        duration=100,
        name='motion',
    )

    # Response Phase

    def _stop(s):
        s.angle_vel = 0.
        s.velocity = np.zeros(2)
    def _unglue(s):
        s.mass = 1.

    appear_cue = gr.ModifySprites('cue', _make_opaque)
    stop_targets = gr.ModifySprites(('targets', 'covers'), _stop)
    unglue_agent = gr.ModifySprites(('agent', 'cue'), _unglue)
    make_targets_discoverable = gr.ModifyOnContact(
        layers_0='agent', layers_1='covers', modifier_1=_make_transparent)

    response_phase = gr.Phase(
        one_time_rules=[appear_cue, stop_targets, unglue_agent],
        continual_rules=make_targets_discoverable,
        name='response',
    )

    phase_sequence = gr.PhaseSequence(
        screen_phase, visible_phase, motion_phase, response_phase,
        meta_state_phase_name_key='phase',
    )

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': (phase_sequence,),
        'meta_state_initializer': lambda: {'phase': ''}
    }
    return config
